; ---------------------------------------------------------------------------
; DQBLITE.ASM
;
; DirectQB: a game programming library for QuickBasic 4.5
;                                 (but it could be easily adapted to your
;                                  own programming language!)
;
; version 1.31 Lite - November, the 23th 1998
;
;     by Angelo Mottola - Enhanced Creations 1998
;
; This is the lite release of DirectQB; it does not contain all the SUBs and
; FUNCTIONs of the full version, but it's quite smaller. See DQBLITE.DOC for
; the complete list of functions of this lite version.
;
; Thanks again for using DirectQB!!
;
; ***************************************************************************

; Let's use a medium memory model...
.MODEL medium,basic

; ...and 386 instructions!
.386

; *** Macros ***

; ---------------------------------------------------------------------------
; WAITWRITEDSP macro
; purpose:
;   Waits the DSP until it's ready to receive data
; calling:
;   DX  Write wait port (BaseAddr+0Ch)
; changes:
;   AL
; ---------------------------------------------------------------------------
WAITWRITEDSP MACRO
LOCAL WaitWrite,EndWriteDSP
  PUSH CX
  XOR CX,CX
WaitWrite:
  DEC CX
  JZ EndWriteDSP
  IN AL,DX
  OR AL,AL
  JS WaitWrite
EndWriteDSP:
  POP CX
ENDM

; ---------------------------------------------------------------------------
; WAITREADDSP macro
; purpose:
;   Waits the DSP until it's ready to send data
; calling:
;   DX  Read wait port (BaseAddr+0Eh)
; changes:
;   AL
; ---------------------------------------------------------------------------
WAITREADDSP MACRO
LOCAL WaitRead,EndReadDSP
  PUSH CX
  XOR CX,CX
WaitRead:
  DEC CX
  JZ EndReadDSP
  IN AL,DX
  OR AL,AL
  JNS WaitRead
EndReadDSP:
  POP CX
ENDM

; ---------------------------------------------------------------------------
; SETUPDMA macro
; purpose:
;   Sets up the DMA chip for the transfer
; ---------------------------------------------------------------------------
SETUPDMA MACRO
LOCAL NoOverRide
  MOV SI,OFFSET SoundBuffer
  MOV CX,255
  MOV AX,@DATA
  ROL AX,4
  MOV BL,AL
  AND BL,00Fh
  AND AL,0F0h
  ADD SI,AX
  ADC BL,0
  NEG SI
  CMP SI,CX
  JA NoOverRide
  NEG SI
  ADD SI,CX
  INC SI
  INC BL
  NEG SI
NoOverRide:
  NEG SI
  MOV AL,DMA
  ADD AL,4
  OUT 0Ah,AL
  OUT 0Ch,AL
  MOV AL,01011000b    ; DMA mode: Single/Autoinit/Read
  ADD AL,DMA
  OUT 0Bh,AL
  MOV AL,BL
  MOV DX,DMA_Page
  OUT DX,AL
  MOV AX,SI
  MOV DX,DMA_Base
  OUT DX,AL
  MOV AL,AH
  OUT DX,AL
  MOV AL,CL
  MOV DX,DMA_Count
  OUT DX,AL
  MOV AL,CH
  OUT DX,AL
  MOV AL,DMA
  OUT 0Ah,AL
  MOV DX,BaseAddr
  ADD DX,0Ch
  WAITWRITEDSP
  MOV AL,040h
  OUT DX,AL
  WAITWRITEDSP
  MOV AL,SamplingRate
  OUT DX,AL
  WAITWRITEDSP
  MOV AL,048h
  OUT DX,AL
  MOV CX,255
  WAITWRITEDSP
  MOV AL,CL
  OUT DX,AL
  WAITWRITEDSP
  MOV AL,CH
  OUT DX,AL
  WAITWRITEDSP
  MOV AL,01Ch
  OUT DX,AL
ENDM

; ---------------------------------------------------------------------------
; MIXVOICES macro
; purpose:
;   Mixes sounds data into the second sound buffer and updates voices
; ---------------------------------------------------------------------------
MIXVOICES MACRO
LOCAL MixSample,UpdateVoices,MixIt,EndMix,MapPage,ClipSound
LOCAL UpdateVoiceLoop,CheckVoiceLoop,NextVoice,LoopVoice
  CLD
  CLI
  MOV AX,@DATA
  MOV ES,AX
  MOV DI,OFFSET MixBuffer
  MOV CX,128
  XOR EAX,EAX     ; Writes complete silence into the sound buffer
  REP STOSD
  CMP Playing,0   ; Avoid mixing if no sounds are playing
  JE EndMix
  XOR CX,CX
MixSample:
  MOV SI,CX
  SHL SI,1
  INC CX
  CMP CX,Channels
  JG UpdateVoices
  CMP VoiceSound[SI],0
  JE MixSample
  PUSH CX
  MOV BX,VoiceSound[SI]
  DEC BX
  SHL BX,1
  MOV CX,SoundLen[BX]
  SHL BX,1
  ADD BX,SoundPage
  MOV DX,EMShdl
  XOR AX,AX       ; Let's clear AX
MapPage2:
  MOV AH,44h      ; LIM/EMS function 44h, starting with physical page 0
  INT 67h         ; Map the page!
  INC BX          ; Next logical page
  INC AL          ; Next physical page
  CMP AL,4        ; Map all the four physical pages
  JNE MapPage2
  MOV AX,EMSseg
  MOV DI,OFFSET MixBuffer
  MOV BX,VoicePos[SI]
  MOV SI,BX
  MOV BX,@DATA
  MOV ES,BX
  MOV DS,AX
  MOV CX,256
MixIt:
  XOR AX,AX
  LODSB
  SUB AX,080h
  ADD ES:[DI],AX
  ADD DI,2
  LOOP MixIt
  POP CX
  MOV AX,@DATA
  MOV DS,AX
  JMP MixSample
UpdateVoices:
  MOV CX,Channels
UpdateVoiceLoop:
  MOV SI,CX
  DEC SI
  SHL SI,1
  MOV DI,VoiceSound[SI]
  CMP DI,0
  JE NextVoice
  CMP VoicePos[SI],0FF00h   ; Has the sound reached the 64K length limit?
  JE CheckVoiceLoop
  DEC DI
  SHL DI,1
  MOV BX,SoundLen[DI]
  CMP VoicePos[SI],BX
  JG CheckVoiceLoop
  ADD VoicePos[SI],256
  LOOP UpdateVoiceLoop
  JMP EndMix
CheckVoiceLoop:
  CMP VoiceLoop[SI],0
  JNE LoopVoice
  MOV VoiceSound[SI],0
  DEC Playing
  LOOP UpdateVoiceLoop
  JMP EndMix
LoopVoice:
  MOV VoicePos[SI],0
NextVoice:
  LOOP UpdateVoiceLoop
EndMix:
  MOV AX,@DATA
  MOV DS,AX
  MOV ES,AX
  MOV SI,OFFSET MixBuffer
  MOV DI,OFFSET SoundBuffer2
  MOV CX,256
CopySample:
  LODSW
  ADD AX,080h
  CMP AX,255
  JLE NoUclip
  MOV AX,255
NoUclip:
  CMP AX,0
  JGE NoLclip
  MOV AX,0
NoLclip:
  STOSB
  LOOP CopySample
  STI
ENDM

; ---------------------------------------------------------------------------
; EPIXEL macro
; purpose:
;   Draws a pixel for the DQBellipse routine
; calling:
;   DX  x
;   BX  y
; ---------------------------------------------------------------------------
EPIXEL MACRO
LOCAL SkipcPixel
  CMP DX,ClipX1
  JL SkipcPixel
  CMP DX,ClipX2
  JG SkipcPixel
  CMP BX,ClipY1
  JL SkipcPixel
  CMP BX,ClipY2
  JG SkipcPixel
  MOV DI,BX
  SHL BX,8
  SHL DI,6
  ADD DI,BX
  ADD DI,DX
  STOSB
SkipcPixel:
ENDM


; 512 bytes of stack are enough for us!
.STACK 200h

; *** Data ***
.DATA

LibVersion    DW  00103h          ; DirectQB library version
EMMid         DB  'EMMXXXX0'      ; EMM id string
EMSseg        DW  ?               ; EMS pageframe segment
EMShdl        DW  ?               ; EMS handle currently being used
EMSpage       DW  ?               ; Current EMS page
EMSallocated  DW  ?               ; Amount of allocated EMS memory
PutMode       DW  0               ; Sprite put mode (0:transparent, 1:solid)
deltax        DW  ?               ; \
deltay        DW  ?               ;  |
d             DW  ?               ;  |
dinc1         DW  ?               ;  |
dinc2         DW  ?               ;   - Bresenham's line algorithm variables
xinc1         DW  ?               ;  |
xinc2         DW  ?               ;  |
yinc1         DW  ?               ;  |
yinc2         DW  ?               ; /
ClipX1        DW  0               ; Left clipping border
ClipX2        DW  319             ; Right clipping border
ClipY1        DW  0               ; Upper clipping border
ClipY2        DW  199             ; Lower clipping border
TextMode      DB  0               ; Text print mode
TextBackCol   DB  0               ; Background text color for solid text mode
FontBuffer    DB  2048 dup (?)    ; Internal font buffer
CharLen       DB  257 dup(8)      ; Internal characters length buffer
Buffer        DB  ?               ; File I/O byte buffer
BSVheader     DB  0FDh,000h,0A0h,000h      ; BSAVE file header (for a 320x200
              DB  000h,000h,0FAh           ; pixels image) - 7 bytes
BMPheader     DB  042h,04Dh,036h,0FEh,000h,000h,000h,000h   ; BMP file header
              DB  000h,000h,036h,004h,000h,000h,028h,000h   ; (for a 320x200
              DB  000h,000h,040h,001h,000h,000h,0C8h,000h   ; pixels image,
              DB  000h,000h,001h,000h,008h,000h,000h,000h   ; uncompressed)
              DB  000h,000h,000h,0FAh,000h,000h,0CEh,00Eh   ; - 54 bytes
              DB  000h,000h,0C4h,00Eh,000h,000h,000h,000h
              DB  000h,000h,000h,000h,000h,000h
PCXheader     DB  00Ah,005h,001h,008h,000h,000h,000h,000h   ; PCX file header
              DB  03Fh,001h,0C7h,000h,040h,001h,0C8h,000h   ; (for a 320x200
              DB  000h,000h,000h,0D8h,098h,038h,078h,074h   ; pixels image,
              DB  004h,070h,06Ch,004h,0ECh,0ACh,04Ch,0F8h   ; encoded using
              DB  0C4h,080h,040h,024h,024h,024h,028h,014h   ; PCX version 3.0
              DB  0F8h,0BCh,068h,0D4h,090h,09Ch,03Ch,024h   ; specifications)
              DB  024h,074h,070h,008h,078h,074h,008h,07Ch   ; - 128 bytes
              DB  078h,008h,034h,030h,004h,0F0h,0C4h,088h
              DB  000h,001h,040h,001h,000h,000h,000h,000h
              DB  000h,000h,000h,000h,000h,000h,000h,000h
              DB  000h,000h,000h,000h,000h,000h,000h,000h
              DB  000h,000h,000h,000h,000h,000h,000h,000h
              DB  000h,000h,000h,000h,000h,000h,000h,000h
              DB  000h,000h,000h,000h,000h,000h,000h,000h
              DB  000h,000h,000h,000h,000h,000h,000h,000h
              DB  000h,000h,000h,000h,000h,000h,000h,000h
KeyActive     DB  0               ; Is the keyboard handler already active?
OldInt9seg    DW  ?               ; Original int 9h segment
OldInt9off    DW  ?               ; Original int 9h offset
KeyFlags      DB  ?               ; Original keyboard flags status
KeyStatus     DB  128 dup(0)      ; Status of each key (0:released, 1:pressed)
SoundActive   DB  0               ; Has the SB custom IRQ been installed?
NumSounds     DB  0               ; Number of allocated sounds
SoundLen      DW  256 dup(0)      ; Sounds length (256 sounds allowed)
SoundPage     DW  ?               ; Starting EMS sound page
BaseAddr      DW  0220h           ; SB base address
IRQ           DB  7               ; SB IRQ
DMA           DB  1               ; SB DMA channel
OldIRQseg     DW  ?               ; Old IRQ interrupt segment
OldIRQoff     DW  ?               ; Old IRQ interrupt offset
PicANDmask    DB  ?               ; AND pic mask to clear IRQ
PicORmask     DB  ?               ; OR pic mask to set IRQ
DMA_Page      DW  ?               ; DMA page register
DMA_Base      DW  ?               ; DMA base address
DMA_Count     DW  ?               ; DMA count register
SamplingRate  DB  165             ; Sampling rate (default = 11000 Hz)
Playing       DB  0               ; Number of sounds currently playing
PauseSound    DB  0               ; Sound pause flag
Channels      DW  8               ; Number of channels used for sample mixing
VoiceSound    DW  32 dup(0)       ; Sounds played by each of the 32 voices
VoicePos      DW  32 dup(0)       ; Sound position of each voice
VoiceLoop     DW  32 dup(0)       ; Loop voice flag (0:play once, 1:loop)
MixBuffer     DW  256 dup(0)      ; Mixing sound buffer
SoundBuffer   DB  256 dup(?)      ; Output sound buffer
SoundBuffer2  DB  256 dup(?)      ; Second sound buffer for double buffering
OldInt24seg   DW  ?               ; Original int 24h segment
OldInt24off   DW  ?               ; Original int 24h offset

; Code begins here!
.CODE


; ***************************************************************************
; Internal library procedures
; ***************************************************************************


; ---------------------------------------------------------------------------
; MapEMS internal function
; purpose:
;   Maps the 4 physical pages of the pageframe beginning from the given
;   logical one, and stores the new page index into EMSpage
; calling:
;   DX  Logical page to begin mapping with
; changes:
;   AX,BX,DX
; ---------------------------------------------------------------------------
MapEMS PROC
  MOV EMSpage,DX  ; Stores the new current page index
  MOV BX,DX
  MOV DX,EMShdl
  XOR AX,AX       ; Let's clear AX
MapPage:
  MOV AH,44h      ; LIM/EMS function 44h, starting with physical page 0
  INT 67h         ; Map the page!
  INC BX          ; Next logical page
  INC AL          ; Next physical page
  CMP AL,4        ; Map all the four physical pages
  JNE MapPage
  RET
MapEMS ENDP

; ---------------------------------------------------------------------------
; GetLayerSeg internal function
; purpose:
;   Sets CX to the current layer segment; if the layer is video, returns
;   A000h, otherwise the EMS pageframe segment.
; calling:
;   BX  layer to retrieve segment of
; changes:
;   CX,BX,DX (,AX if MAPEMS is called)
; ---------------------------------------------------------------------------
GetLayerSeg PROC
  MOV CX,0A000h
  CMP BX,0        ; Is specified layer the video?
  JNE LayerOnEMS  ; No: layer is on EMS
  RET             ; Layer is video: returns screen segment (A000h)
LayerOnEMS:
  MOV CX,EMSseg   ; Returns the EMS pageframe segment
  DEC BX
  SHL BX,2        ; Translates the layer number into an EMS page index
  CMP BX,EMSpage  ; Do we need to change the actually mapped EMS page?
  JNE OnDifferentPage   ; Yes: map the new page
  RET
OnDifferentPage:
  MOV DX,BX
  CALL MapEMS          ; Calls the MAPEMS procedure to map the new page
  MOV CX,EMSseg   ; Returns the EMS pageframe segment (restored)
  RET
GetLayerSeg ENDP

; ---------------------------------------------------------------------------
; WriteBuffer internal procedure
; purpose:
;   Writes the buffer to file
; ---------------------------------------------------------------------------
WriteBuffer PROC
  PUSH DS
  PUSH AX
  PUSH CX
  PUSH DX
  MOV DX,OFFSET Buffer
  MOV AX,@DATA
  MOV DS,AX
  MOV CX,1
  MOV AH,40h
  INT 21h
  POP DX
  POP CX
  POP AX
  POP DS
  RET
WriteBuffer ENDP

; ---------------------------------------------------------------------------
; ReadBuffer internal procedure
; purpose:
;   Reads data from file and places it into the buffer
; ---------------------------------------------------------------------------
ReadBuffer PROC
  PUSH DS
  PUSH AX
  PUSH CX
  PUSH DX
  MOV DX,OFFSET Buffer
  MOV AX,@DATA
  MOV DS,AX
  MOV CX,1
  MOV AH,3Fh
  INT 21h
  POP DX
  POP CX
  POP AX
  POP DS
  RET
ReadBuffer ENDP

; ---------------------------------------------------------------------------
; HookInt24 internal procedure
; purpose:
;   Installs custom INT 24h handler
; ---------------------------------------------------------------------------
HookInt24 PROC
  PUSH DS
  MOV AX,3524h
  INT 21h
  MOV AX,@DATA
  MOV DS,AX
  MOV OldInt24seg,ES
  MOV OldInt24off,BX
  POP DS
  MOV AX,SEG CriticalISR
  MOV DX,OFFSET CriticalISR
  PUSH DS
  MOV DS,AX
  MOV AX,2524h
  INT 21h
  POP DS
  RET
CriticalISR:
  MOV AL,3
  IRET
HookInt24 ENDP

; ---------------------------------------------------------------------------
; UnhookInt24 internal procedure
; purpose:
;   Removes the custom int 24h handler
; ---------------------------------------------------------------------------
UnhookInt24 PROC
  PUSH AX
  PUSH DS
  MOV AX,@DATA
  MOV DS,AX
  MOV DX,OldInt24off
  MOV AX,OldInt24seg
  MOV DS,AX
  MOV AX,2524h
  INT 21h
  POP DS
  POP AX
  RET
UnhookInt24 ENDP


; ***************************************************************************
; GENERAL PURPOSE ROUTINES
; ***************************************************************************


; ---------------------------------------------------------------------------
; DQBinit FUNCTION
; purpose:
;   Initializes the library, by allocating the needed amount of EMS memory
;   to store layers (if requested). Also checks if a 386 CPU is available on
;   the system, calibrates the joysticks (where available) and sets up the
;   mouse. On success returns 0, otherwise an error code.
;   Always call this procedure before calling any of the library functions!
; declaration:
;   DECLARE FUNCTION DQBinit(BYVAL NumLayers,BYVAL NumSounds)
; ---------------------------------------------------------------------------
PUBLIC DQBinit
DQBinit PROC
  ; Stack layout:
  ;
  ; 08  NumLayers
  ; 06  NumSounds
  ; 04  Basic segment
  ; 02  Basic offset
  ; 00  BP
  PUSH BP
  MOV BP,SP
  PUSH DS
  MOV AX,@DATA
  MOV ES,AX
  MOV DI,OFFSET FontBuffer
  MOV AX,0FFA6h
  MOV DS,AX
  MOV SI,0Eh
  MOV CX,512
  REP MOVSD
  MOV EAX,08080808h
  MOV CX,64
  MOV DI,OFFSET CharLen
  REP STOSD
  STOSB
  POP DS
  PUSHF
  MOV CX,7000h    ; Set bits 12..15
  PUSH CX
  POPF
  PUSHF           ; Push them again...
  POP AX          ; ...and get them back
  AND AX,7000h    ; Mask bits 12..15
  SETNZ AL
  XOR AH,AH       ; Zero it
  POPF
  CMP AX,0
  JE Error_NoCPU
  XOR AX,AX
  MOV EMSallocated,AX
  MOV DX,[BP+8]
  SHL DX,2
  MOV SoundPage,DX
  MOV DX,[BP+6]
  MOV NumSounds,DL
  MOV DX,[BP+8]   ; Let's find out the total amount of EMS needed
  ADD DX,[BP+6]
  SHL DX,2        ; EMS needed=(NumLayers+NumSounds)*4 pages
  CMP DX,0
  JE MemDone      ; no EMS memory is needed, so skip memory allocation
  MOV ES,AX
  MOV BX,19Eh     ; Int 67h address
  MOV AX,ES:[BX]  ; Gets the EMM segment from the Interrupts Vector Table
  MOV ES,AX       ; ES now is the segment were the EMM driver is loaded
  MOV CL,8        ; Let's examine the 8 characters of the EMM id
  MOV SI,10       ; ES:SI now points to the supposed EMM id
  XOR BX,BX
CmpLoop:
  MOV AL,ES:[SI]
  CMP AL,DS:[EMMid+BX]    ; Compares each character
  JNE Error_NoEMS         ; If one of them is not the supposed one, fails
  INC SI          ; \
  INC BX          ;  |- Process next character
  DEC CL          ; /
  JNZ CmpLoop
  MOV AH,42h
  INT 67h         ; Let's check if enough EMS is available
  MOV DX,[BP+8]   ; Let's find out again the total amount of EMS needed
  ADD DX,[BP+6]
  SHL DX,2        ; EMS needed=(NumLayers+NumSounds)*4 pages
  CMP BX,DX       ; BX contains the number of available pages
  JL Error_NotEnoughEMS     ; not enough memory!
  MOV EMSallocated,DX       ; Stores amount of allocated EMS pages
  MOV AH,43h
  MOV BX,DX
  INT 67h         ; Allocates memory
  MOV EMShdl,DX   ; Stores the EMS handle
  MOV AH,41h
  INT 67h         ; Gets the EMS pageframe...
  MOV EMSseg,BX   ; ...and stores it
  MOV ES,BX
  MOV BX,[BP+6]
  CMP BX,0
  JE CheckLayers
  MOV EAX,080808080h
  ADD BX,[BP+8]
InitSounds:
  MOV DX,BX
  PUSH AX
  PUSH BX
  PUSH DX
  CALL MapEMS
  POP DX
  POP BX
  POP AX
  XOR DI,DI
  MOV CX,16384
  REP STOSD
  DEC BX
  JNZ InitSounds
CheckLayers:
  MOV BX,[BP+8]
  CMP BX,0
  JE MemDone
  XOR EAX,EAX
InitLayers:
  MOV DX,BX
  PUSH AX
  PUSH BX
  PUSH DX
  CALL MapEMS
  POP DX
  POP BX
  POP AX
  XOR DI,DI
  MOV CX,16384
  REP STOSD
  DEC BX
  JNZ InitLayers
  XOR DX,DX
  CALL MapEMS     ; Maps the pageframe starting from logical page 0
  XOR AX,AX       ; EMS memory initialized!
  JMP MemDone
Error_NoCPU:
  MOV AX,1        ; Error 1: No 386 CPU detected!
  JMP EndInit
Error_NoEMS:
  MOV AX,2        ; Error 2: No EMS memory manager detected!
  JMP EndInit
Error_NotEnoughEMS:
  MOV AX,3        ; Error 3: Not enough free EMS memory!
  JMP EndInit
MemDone:
  XOR AX,AX
EndInit:
  POP BP
  RET 4
DQBinit ENDP

; ---------------------------------------------------------------------------
; DQBver FUNCTION
; purpose:
;   Returns the library version number in hexadecimal format; the higher byte
;   is the major release number, and the lower byte is the minor release
;   number.
; declaration:
;   DECLARE FUNCTION DQBver()
; ---------------------------------------------------------------------------
PUBLIC DQBver
DQBver PROC
  MOV AX,LibVersion
  RET
DQBver ENDP

; ---------------------------------------------------------------------------
; DQBemsseg FUNCTION
; purpose:
;   Returns the EMS pageframe segment
; declaration:
;   DECLARE FUNCTION DQBemsseg()
; ---------------------------------------------------------------------------
PUBLIC DQBemsseg
DQBemsseg PROC
  MOV AX,EMSseg   ; AX (function return value) holds the EMS pageframe
  RET
DQBemsseg ENDP

; ---------------------------------------------------------------------------
; DQBclose SUB
; purpose:
;   Deallocates previously allocated EMS memory (if it was allocated), and
;   does other closing stuff needed.
;   ALWAYS CALL THIS SUB BEFORE ENDING YOUR PROGRAMS!
; declaration:
;   DECLARE SUB DQBclose()
; ---------------------------------------------------------------------------
PUBLIC DQBclose
DQBclose PROC
  CMP KeyActive,0
  JE NoRemoveKey
  PUSH DS
  MOV DX,OldInt9off
  MOV AX,OldInt9seg
  MOV DS,AX
  MOV AX,2509h
  INT 21h
  POP DS
  MOV BL,KeyFlags
  MOV AX,040h
  MOV ES,AX
  MOV SI,017h
  MOV ES:[SI],BL
  MOV KeyActive,0
NoRemoveKey:
  CMP SoundActive,0
  JE NoRemoveSB
  MOV DX,BaseAddr
  ADD DX,06h
  MOV AL,1
  OUT DX,AL
  IN AL,DX
  IN AL,DX
  IN AL,DX
  IN AL,DX
  XOR AL,AL
  OUT DX,AL
  ADD DX,08h
  WAITREADDSP
  SUB DX,4
  IN AL,DX          ; SB reset
  IN AL,021h
  OR AL,PicORmask
  OUT 021h,AL
  XOR AX,AX
  MOV ES,AX
  MOV AL,IRQ
  ADD AL,8
  MOV SI,AX
  SHL SI,2
  CLI
  MOV AX,OldIRQoff
  MOV ES:[SI],AX
  MOV AX,OldIRQseg
  MOV ES:[SI+2],AX  ; Old IRQ restored
  STI
  MOV SoundActive,0
NoRemoveSB:
  CMP EMSallocated,0    ; Did we allocate EMS memory?
  JE NoDealloc      ; No EMS memory was allocated at startup!
  MOV AH,45h        ; Function 45h: release allocated EMS memory
  MOV DX,EMShdl
  INT 67h
NoDealloc:
  RET
DQBclose ENDP


; ***************************************************************************
; GRAPHICAL ROUTINES
; ***************************************************************************

; ---------------------------------------------------------------------------
; DQBInitVGA SUB
; purpose:
;   Sets VGA 320x200 with 256 colors, and initializes the mouse if found.
; declaration:
;   DECLARE SUB DQBinitVGA
; ---------------------------------------------------------------------------
PUBLIC DQBinitVGA
DQBinitVGA PROC
  MOV AX,013h
  INT 010h
  RET
DQBinitVGA ENDP

; ---------------------------------------------------------------------------
; DQBInitText SUB
; purpose:
;   Sets plain 80x25 text mode.
; declaration:
;   DECLARE SUB DQBinitText()
; ---------------------------------------------------------------------------
PUBLIC DQBinitText
DQBinitText PROC
  MOV AX,03h
  INT 010h
  RET
DQBinitText ENDP

; ---------------------------------------------------------------------------
; DQBcopyLayer SUB
; purpose:
;   Copies a layer onto another one.
; declaration:
;   DECLARE SUB DQBcopyLayer(BYVAL SourceLayer,BYVAL DestLayer)
; ---------------------------------------------------------------------------
PUBLIC DQBcopyLayer
DQBcopyLayer PROC
  ; Stack layout:
  ;
  ; 10  SourceLayer
  ; 08  DestLayer
  ; 06  Basic segment
  ; 04  Basic offset
  ; 02  DS
  ; 00  BP
  PUSH DS
  PUSH BP
  MOV BP,SP
  MOV BX,[BP+10]  ; BX holds the source layer number
  MOV DX,[BP+8]   ; DX holds the destination layer number
  CMP BX,DX       ; Are we processing the same layers?
  JE EndCopy      ; Yes: exits function
  CMP BX,0
  JE OneIsVideo   ; Is source layer on video?
  CMP DX,0
  JE OneIsVideo   ; Is destination layer on video?
  MOV DX,EMShdl   ; Both source and destination layer are on EMS
  MOV AX,EMSseg
  MOV DS,AX       ; \
  MOV ES,AX       ; |- Both DS and ES points to the EMS pageframe segment
  DEC BX
  SHL BX,2
  MOV AX,4400h
  INT 67h
  MOV AX,4401h
  INC BX
  INT 67h
  MOV BX,[BP+8]
  DEC BX
  SHL BX,2
  MOV AX,4402h
  INT 67h
  MOV AX,4403h    ; At the end the first two physical pages are mapped to the    
  INC BX          ; first half of the source layer, and the third and fourth
  INT 67h         ; pages points to the first half of the destination layer
  MOV SI,0
  MOV DI,8000h
  MOV CX,2000h    ; Copy the DWORDs!
  REP MOVSD
  MOV AX,4400h
  MOV BX,[BP+10]
  DEC BX
  SHL BX,2
  ADD BX,2
  INT 67h
  MOV AX,4401h
  INC BX
  INT 67h
  MOV BX,[BP+8]
  DEC BX
  SHL BX,2
  ADD BX,2
  MOV AX,4402h
  INT 67h
  MOV AX,4403h
  INC BX
  INT 67h
  MOV SI,0
  MOV DI,8000h
  MOV CX,2000h
  REP MOVSD
  MOV AX,@DATA
  MOV DS,AX
  MOV DX,EMSpage
  CALL MapEMS
  JMP EndCopy
OneIsVideo:
  CALL GetLayerSeg
  MOV DI,CX       ; Saves source segment into DI
  MOV BX,[BP+8]
  CALL GetLayerSeg
  MOV ES,CX       ; Sets destination segment...
  MOV DS,DI       ; ...and source segment
  MOV DI,0        ; Both source and destination offsets start as 0
  MOV SI,0
  MOV CX,16000    ; Move 16000 DWORDs
  REP MOVSD
EndCopy:
  POP BP
  POP DS
  RET 4
DQBcopyLayer ENDP

; ---------------------------------------------------------------------------
; DQBclearLayer SUB
; purpose:
;   Clears the content of a given layer to black (color 0 used)
; declaration:
;   DECLARE SUB DQBclearLayer(BYVAL Layer)
; ---------------------------------------------------------------------------
PUBLIC DQBclearLayer
DQBclearLayer PROC
  ; Stack layout:
  ;
  ; 06  Layer
  ; 04  Basic return segment
  ; 02  Basic return offset
  ; 00  BP
  PUSH BP
  MOV BP,SP
  MOV BX,[BP+6]
  CALL GetLayerSeg
  MOV ES,CX
  MOV CX,16384
  XOR EAX,EAX
  XOR DI,DI
  REP STOSD
  POP BP
  RET 2
DQBclearLayer ENDP

; ---------------------------------------------------------------------------
; DQBpset SUB
; purpose:
;   Writes a pixel on the given segment, assuming an offset of 0 for the first
;   pixel on the upper-left corner of the screen.
; declaration:
;   DECLARE SUB DQBpset(BYVAL Layer,BYVAL x,BYVAL y,BYVAL Col)
; ---------------------------------------------------------------------------
PUBLIC DQBpset
DQBpset PROC
  ; Stack layout:
  ; 
  ; 12  Layer
  ; 10  x
  ; 08  y
  ; 06  Color
  ; 04  Basic segment
  ; 02  Basic offset
  ; 00  BP
  PUSH BP
  MOV BP,SP
  MOV AX,[BP+10]
  CMP AX,ClipX1
  JL ExitPset
  CMP AX,ClipX2
  JG ExitPset
  MOV AX,[BP+8]
  CMP AX,ClipY1
  JL ExitPset
  CMP AX,ClipY2
  JG ExitPset
  MOV BX,[BP+12]    ; Gets the output layer
  CALL GetLayerSeg
  MOV ES,CX         ; Sets the appropriate segment
  MOV DI,[BP+8]    ; Calculates the offset
  MOV BX,DI
  SHL DI,8
  SHL BX,6
  ADD DI,BX
  ADD DI,[BP+10]
  MOV AL,[BP+6]     ; Gets the pixel color...
  STOSB
ExitPset:
  POP BP
  RET 8
DQBpset ENDP

; ---------------------------------------------------------------------------
; DQBpoint FUNCTION
; purpose:
;   Returns the color of the pixel at the given coordinates and on the given
;   layer.
; declaration:
;   FUNCTION DQBpoint(BYVAL Layer,BYVAL x,BYVAL y)
; ---------------------------------------------------------------------------
PUBLIC DQBpoint
DQBpoint PROC
  ; Stack layout:
  ; 
  ; 10  Layer
  ; 08  x
  ; 06  y
  ; 04  Basic segment
  ; 02  Basic offset
  ; 00  BP
  PUSH BP
  MOV BP,SP
  MOV BX,[BP+10]    ; Gets the output layer
  CALL GetLayerSeg
  MOV ES,CX         ; Sets the appropriate segment
  MOV SI,[BP+6]     ; Calculates the offset
  MOV BX,SI
  SHL SI,8
  SHL BX,6
  ADD SI,BX
  ADD SI,[BP+8]
  XOR AH,AH
  MOV AL,ES:[SI]
  POP BP
  RET 6
DQBpoint ENDP

; ---------------------------------------------------------------------------
; DQBline SUB
; purpose:
;   Draw a line from x1,y1 to x2,y2 on specified layer with the given color,
;   using the Bresenham's line algorithm.
; declaration:
;   DECLARE SUB DQBline(BYVAL Layer,BYVAL x1,BYVAL y1,BYVAL x2,BYVAL y2,
;                       BYVAL Col)
; ---------------------------------------------------------------------------
PUBLIC DQBline
DQBline PROC
  ; Stack layout:
  ;
  ; 16  Layer
  ; 14  x1
  ; 12  y1
  ; 10  x2
  ; 08  y2
  ; 06  Col
  ; 04  Basic return segment
  ; 02  Basic return offset
  ; 00  BP
  PUSH BP
  MOV BP,SP
  MOV BX,[BP+16]
  CALL GetLayerSeg
  MOV ES,CX
  MOV AX,[BP+14]
  MOV BX,[BP+10]
  CMP AX,BX
  JG Line1
  SUB BX,AX
  MOV deltax,BX
  JMP Line2
Line1:
  SUB AX,BX
  MOV deltax,AX
Line2:
  MOV AX,[BP+12]
  MOV BX,[BP+8]
  CMP AX,BX
  JG Line3
  SUB BX,AX
  MOV deltay,BX
  JMP Line4
Line3:
  SUB AX,BX
  MOV deltay,AX
Line4:
  MOV AX,deltax
  CMP AX,deltay
  JL Line5
  MOV CX,deltax
  INC CX
  MOV BX,deltay
  SHL BX,1
  MOV dinc1,BX
  SUB BX,AX
  MOV d,BX
  MOV BX,deltay
  SUB BX,AX
  SHL BX,1
  MOV dinc2,BX
  MOV xinc1,1
  MOV xinc2,1
  MOV yinc1,0
  MOV yinc2,1
  JMP Line6
Line5:
  MOV CX,deltay
  INC CX
  MOV BX,deltax
  SHL BX,1
  MOV dinc1,BX
  SUB BX,CX
  MOV d,BX
  SUB AX,CX
  INC AX
  SHL AX,1
  MOV dinc2,AX
  MOV xinc1,0
  MOV xinc2,1
  MOV yinc1,1
  MOV yinc2,1
Line6:
  MOV AX,[BP+14]
  CMP AX,[BP+10]
  JLE Line7
  MOV DX,xinc1
  MOV AX,-1
  MUL DX
  MOV xinc1,AX
  MOV DX,xinc2
  MOV AX,-1
  MUL DX
  MOV xinc2,AX
Line7:
  MOV AX,[BP+12]
  CMP AX,[BP+8]
  JLE Line8
  MOV DX,yinc1
  MOV AX,-1
  MUL DX
  MOV yinc1,AX
  MOV DX,yinc2
  MOV AX,-1
  MUL DX
  MOV yinc2,AX
Line8:
  MOV DX,[BP+14]    ; DX holds the current x coordinate
  MOV BX,[BP+12]    ; BX holds the current y coordinate
  MOV AX,[BP+6]     ; AL holds the color
LineLoop:
  CMP DX,ClipX1
  JL SkipLinePixel
  CMP DX,ClipX2
  JG SkipLinePixel
  CMP BX,ClipY1
  JL SkipLinePixel
  CMP BX,ClipY2
  JG SkipLinePixel
  MOV DI,BX
  MOV SI,BX
  SHL DI,8
  SHL SI,6
  ADD DI,SI
  ADD DI,DX
  STOSB
SkipLinePixel:
  CMP d,0
  JL Line9
  MOV SI,d
  ADD SI,dinc2
  MOV d,SI
  ADD DX,xinc2
  ADD BX,yinc2
  JMP Line10
Line9:
  MOV SI,d
  ADD SI,dinc1
  MOV d,SI
  ADD DX,xinc1
  ADD BX,yinc1
Line10:
  LOOP LineLoop
  POP BP
  RET 12
DQBline ENDP

; ---------------------------------------------------------------------------
; DQBellipse SUB
; purpose:
;   Draws an ellipse on specified layer with center in (x,y), and rx,ry as
;   the horizontal and vertical radiuses, with col color. The function works
;   for any radiuses ranging 0-255, otherwise you'll get unwanted results.
;   DQBellipse is affected by the clipping box.
; declaration:
;   DECLARE SUB DQBellipse(BYVAL Layer,BYVAL x,BYVAL y,BYVAL rx,BYVAL ry,
;                          BYVAL col)
; ---------------------------------------------------------------------------
PUBLIC DQBellipse
DQBellipse PROC
  ; Stack layout:
  ;
  ; 28  Layer
  ; 26  x
  ; 24  y
  ; 22  rx
  ; 20  ry
  ; 18  Col
  ; 16  Basic return segment
  ; 14  Basic return offset
  ; 12  BP
  ; 10  x
  ; 08  y
  ; 06  xe
  ; 04  ye
  ; 02  e
  ; 00  c
  PUSH BP
  MOV BP,SP
  MOV BX,[BP+16]
  CALL GetLayerSeg
  MOV ES,CX
  PUSH AX           ; Preserve stack space for internal variables
  PUSH AX
  PUSH AX
  PUSH AX
  PUSH AX
  PUSH AX
  MOV BP,SP
  MOV AX,[BP+22]
  CMP AX,0
  JNE NotVline
  MOV AX,[BP+18]
  MOV CX,[BP+20]
  NEG CX
eVline:
  MOV DX,[BP+26]
  MOV BX,[BP+24]
  ADD BX,CX
  EPIXEL
  INC CX
  CMP CX,[BP+20]
  JLE eVline
  JMP EndEllipse
NotVline:
  MOV AX,[BP+20]
  CMP AX,0
  JNE NotHline
  MOV AX,[BP+18]
  MOV CX,[BP+22]
  NEG CX
eHline:
  MOV DX,[BP+26]
  ADD DX,CX
  MOV BX,[BP+24]
  EPIXEL
  INC CX
  CMP CX,[BP+22]
  JLE eHline
  JMP EndEllipse
NotHline:
  CMP AX,[BP+22]
  JG OtherAxis
  XOR AX,AX
  MOV [BP+10],AX
  MOV [BP+6],AX
  MOV AX,[BP+20]
  MOV [BP+8],AX
  MOV AX,[BP+22]
  MOV BX,AX
  MUL BX
  MOV [BP+4],AX
  PUSH AX
  SHR AX,1
  NEG AX
  MOV [BP+2],AX
  POP AX
  XOR DX,DX
  MOV BX,[BP+20]
  DIV BX
  MOV [BP],AX
  MOV AX,[BP+18]
Circle1:
  MOV CX,[BP+2]
  CMP CX,0
  JG Circle3
Circle2:
  MOV DX,[BP+26]
  ADD DX,[BP+10]
  MOV BX,[BP+24]
  ADD BX,[BP+8]
  EPIXEL
  MOV DX,[BP+26]
  SUB DX,[BP+10]
  MOV BX,[BP+24]
  ADD BX,[BP+8]
  EPIXEL
  MOV DX,[BP+26]
  ADD DX,[BP+10]
  MOV BX,[BP+24]
  SUB BX,[BP+8]
  EPIXEL
  MOV DX,[BP+26]
  SUB DX,[BP+10]
  MOV BX,[BP+24]
  SUB BX,[BP+8]
  EPIXEL
  MOV BX,[BP+10]
  INC BX
  MOV [BP+10],BX
  MOV BX,[BP+6]
  ADD BX,[BP+20]
  MOV [BP+6],BX
  MOV BX,[BP+2]
  ADD BX,[BP+6]
  MOV [BP+2],BX
  CMP BX,0
  JLE Circle2
  JMP Circle4
Circle3:
  MOV DX,[BP+26]
  ADD DX,[BP+10]
  MOV BX,[BP+24]
  ADD BX,[BP+8]
  EPIXEL
  MOV DX,[BP+26]
  SUB DX,[BP+10]
  MOV BX,[BP+24]
  ADD BX,[BP+8]
  EPIXEL
  MOV DX,[BP+26]
  ADD DX,[BP+10]
  MOV BX,[BP+24]
  SUB BX,[BP+8]
  EPIXEL
  MOV DX,[BP+26]
  SUB DX,[BP+10]
  MOV BX,[BP+24]
  SUB BX,[BP+8]
  EPIXEL
Circle4:
  MOV BX,[BP+4]
  SUB BX,[BP]
  MOV [BP+4],BX
  MOV BX,[BP+2]
  SUB BX,[BP+4]
  MOV [BP+2],BX
  MOV BX,[BP+8]
  DEC BX
  MOV [BP+8],BX
  CMP BX,0
  JNZ Circle1
  MOV DX,[BP+26]
  ADD DX,[BP+10]
  MOV BX,[BP+24]
  EPIXEL
  MOV DX,[BP+26]
  SUB DX,[BP+10]
  MOV BX,[BP+24]
  EPIXEL
  JMP EndEllipse
OtherAxis:
  XOR AX,AX
  MOV [BP+10],AX
  MOV [BP+6],AX
  MOV AX,[BP+22]
  MOV [BP+8],AX
  MOV AX,[BP+20]
  MOV BX,AX
  MUL BX
  MOV [BP+4],AX
  PUSH AX
  SHR AX,1
  NEG AX
  MOV [BP+2],AX
  POP AX
  XOR DX,DX
  MOV BX,[BP+22]
  DIV BX
  MOV [BP],AX
  MOV AX,[BP+18]
Circle5:
  MOV CX,[BP+2]
  CMP CX,0
  JG Circle7
Circle6:
  MOV DX,[BP+26]
  ADD DX,[BP+8]
  MOV BX,[BP+24]
  ADD BX,[BP+10]
  EPIXEL
  MOV DX,[BP+26]
  SUB DX,[BP+8]
  MOV BX,[BP+24]
  ADD BX,[BP+10]
  EPIXEL
  MOV DX,[BP+26]
  ADD DX,[BP+8]
  MOV BX,[BP+24]
  SUB BX,[BP+10]
  EPIXEL
  MOV DX,[BP+26]
  SUB DX,[BP+8]
  MOV BX,[BP+24]
  SUB BX,[BP+10]
  EPIXEL
  MOV BX,[BP+10]
  INC BX
  MOV [BP+10],BX
  MOV BX,[BP+6]
  ADD BX,[BP+22]
  MOV [BP+6],BX
  MOV BX,[BP+2]
  ADD BX,[BP+6]
  MOV [BP+2],BX
  CMP BX,0
  JLE Circle6
  JMP Circle8
Circle7:
  MOV DX,[BP+26]
  ADD DX,[BP+8]
  MOV BX,[BP+24]
  ADD BX,[BP+10]
  EPIXEL
  MOV DX,[BP+26]
  SUB DX,[BP+8]
  MOV BX,[BP+24]
  ADD BX,[BP+10]
  EPIXEL
  MOV DX,[BP+26]
  ADD DX,[BP+8]
  MOV BX,[BP+24]
  SUB BX,[BP+10]
  EPIXEL
  MOV DX,[BP+26]
  SUB DX,[BP+8]
  MOV BX,[BP+24]
  SUB BX,[BP+10]
  EPIXEL
Circle8:
  MOV BX,[BP+4]
  SUB BX,[BP]
  MOV [BP+4],BX
  MOV BX,[BP+2]
  SUB BX,[BP+4]
  MOV [BP+2],BX
  MOV BX,[BP+8]
  DEC BX
  MOV [BP+8],BX
  CMP BX,0
  JNZ Circle5
  MOV DX,[BP+26]
  MOV BX,[BP+24]
  ADD BX,[BP+10]
  EPIXEL
  MOV DX,[BP+26]
  MOV BX,[BP+24]
  SUB BX,[BP+10]
  EPIXEL
EndEllipse:
  POP AX
  POP AX
  POP AX
  POP AX
  POP AX
  POP AX
  POP BP
  RET 12
DQBellipse ENDP

; ---------------------------------------------------------------------------
; DQBbox SUB
; purpose:
;   Draws an empty box with x1,y1 and x2,y2 as the opposite corners with a
;   given color. Please note that must be x1<x2 and y1<y2; no checks are
;   done, and clipping is not supported to speed things up.
; declaration:
;   DECLARE SUB DQBbox(BYVAL Layer,BYVAL x1,BYVAL y1,BYVAL x2,BYVAL y2,
;                       BYVAL Col)
; ---------------------------------------------------------------------------
PUBLIC DQBbox
DQBbox PROC
  ; Stack layout:
  ;
  ; 16  Layer
  ; 14  x1
  ; 12  y1
  ; 10  x2
  ; 08  y2
  ; 06  Col
  ; 04  Basic return segment
  ; 02  Basic return offset
  ; 00  BP
  PUSH BP
  MOV BP,SP
  MOV BX,[BP+16]
  CALL GetLayerSeg
  MOV ES,CX
  MOV DI,[BP+12]
  MOV BX,DI
  SHL DI,8
  SHL BX,6
  ADD DI,BX
  ADD DI,[BP+14]
  MOV CX,[BP+10]
  SUB CX,[BP+14]
  INC CX
  MOV AL,[BP+6]
  MOV AH,AL
  MOV DX,CX
  PUSH CX
  CMP CX,2
  JL BoxByteTest1
  SHR CX,1
  REP STOSW
BoxByteTest1:
  TEST DX,1
  JZ DrawBorders
  STOSB
DrawBorders:
  POP DX
  SUB DX,2
  MOV DI,[BP+12]
  INC DI
  MOV BX,DI
  SHL DI,8
  SHL BX,6
  ADD DI,BX
  ADD DI,[BP+14]
  MOV CX,[BP+8]
  SUB CX,[BP+12]
  DEC CX
  PUSH CX
  CMP CX,0
  JLE LowerBorder
BorderLoop:
  STOSB
  ADD DI,DX
  STOSB
  ADD DI,318
  SUB DI,DX
  LOOP BorderLoop
LowerBorder:
  POP CX
  CMP CX,0
  JL EndBox
  MOV CX,[BP+10]
  SUB CX,[BP+14]
  INC CX
  MOV DX,CX
  CMP CX,2
  JL BoxByteTest2
  SHR CX,1
  REP STOSW
BoxByteTest2:
  TEST DX,1
  JZ EndBox
  STOSB
EndBox:
  POP BP
  RET 12
DQBbox ENDP

; ---------------------------------------------------------------------------
; DQBboxf SUB
; purpose:
;   Draws an full box with x1,y1 and x2,y2 as the opposite corners with a
;   given color. Please note that must be x1<x2 and y1<y2; no checks are
;   done and clipping is not supported to speed things up.
; declaration:
;   DECLARE SUB DQBboxf(BYVAL Layer,BYVAL x1,BYVAL y1,BYVAL x2,BYVAL y2,
;                       BYVAL Col)
; ---------------------------------------------------------------------------
PUBLIC DQBboxf
DQBboxf PROC
  ; Stack layout:
  ;
  ; 16  Layer
  ; 14  x1
  ; 12  y1
  ; 10  x2
  ; 08  y2
  ; 06  Col
  ; 04  Basic return segment
  ; 02  Basic return offset
  ; 00  BP
  PUSH BP
  MOV BP,SP
  MOV BX,[BP+16]
  CALL GetLayerSeg
  MOV ES,CX
  MOV DI,[BP+12]
  MOV BX,DI
  SHL DI,8
  SHL BX,6
  ADD DI,BX
  ADD DI,[BP+14]
  MOV CX,[BP+10]
  SUB CX,[BP+14]
  INC CX
  MOV DX,[BP+8]
  SUB DX,[BP+12]
  INC DX
  MOV AL,[BP+6]
  MOV AH,AL
BoxfLoop:
  PUSH CX
  SHR CX,1
  CMP CX,0
  JE BoxfByteTest
  REP STOSW
BoxfByteTest:
  POP CX
  TEST CX,1
  JZ NextBoxfLine
  STOSB
NextBoxfLine:
  ADD DI,320
  SUB DI,CX
  DEC DX
  JNZ BoxfLoop
  POP BP
  RET 12
DQBboxf ENDP

; ---------------------------------------------------------------------------
; DQBscroll SUB
; purpose:
;   Scrolls specified layer with dx and dy increments. It's up to the
;   programmer to draw the layer area that should be replaced.
; declaration:
;   DECLARE SUB DQBscroll(BYVAL Layer,BYVAL dx,BYVAL dy)
; ---------------------------------------------------------------------------
PUBLIC DQBscroll
DQBscroll PROC
  ; Stack layout:
  ;
  ; 12    Layer
  ; 10    dx
  ; 08    dy
  ; 06    Basic return segment
  ; 04    Basic return offset
  ; 02    DS
  ; 00    BP
  PUSH DS
  PUSH BP
  MOV BP,SP
  MOV BX,[BP+12]
  CALL GetLayerSeg  ; \
  MOV DS,CX         ;  |- Sets both source and destination segments to the
  MOV ES,CX         ; /   specified layer segment
  MOV AX,[BP+08]
  MOV BX,320
  IMUL BX
  SHL EDX,16
  MOV DX,AX         ; Now EDX holds 320*dy
  MOV AX,[BP+10]
  CWDE
  ADD EDX,EAX       ; EDX holds the total number of bytes to shift (signed)
  OR EDX,EDX
  JNS Positive
  CLD
  XOR SI,SI
  MOV DI,DX
  MOV CX,64000
  SUB CX,DX         ; CX holds the number of bytes to copy
  JMP Scroll
Positive:
  STD               ; Sets the direction flag
  MOV CX,64000
  SUB CX,DX
  MOV SI,CX
  INC CX            ; CX holds the number of bytes to copy
  MOV DI,64000
Scroll:
  REP MOVSB         ; Moves the bytes
  CLD               ; Clears the direction flag
  POP BP
  POP DS
  RET 6
DQBscroll ENDP

; ---------------------------------------------------------------------------
; DQBwait
; purpose:
;   Waits for video vertical retrace as many times as specified when called
; declaration:
;   DECLARE SUB DQBwait(BYVAL times)
; ---------------------------------------------------------------------------
PUBLIC DQBwait
DQBwait PROC
  PUSH BP
  MOV BP,SP
  MOV CX,[BP+6]
  MOV DX,3DAh       ; Vertical retrace port
Loop1:
  IN AL,DX
  AND AL,8
  JNZ Loop1
Loop2:
  IN AL,DX
  AND AL,8
  JZ Loop2
  LOOP Loop1
  POP BP
  RET 2
DQBwait ENDP

; ---------------------------------------------------------------------------
; DQBsetTransPut SUB
; purpose:
;   Sets the transparent sprite put mode. Next DQBput calls will draw
;   transparent sprites.
; declaration:
;   DECLARE SUB DQBsetTransPut()
; ---------------------------------------------------------------------------
PUBLIC DQBsetTransPut
DQBsetTransPut PROC
  MOV PutMode,0
  RET
DQBsetTransPut ENDP

; ---------------------------------------------------------------------------
; DQBsetSolidPut SUB
; purpose:
;   Sets the solid sprite put mode. Next DQBput calls will draw solid sprites.
; declaration:
;   DECLARE SUB DQBsetSolidPut()
; ---------------------------------------------------------------------------
PUBLIC DQBsetSolidPut
DQBsetSolidPut PROC
  MOV PutMode,1
  RET
DQBsetSolidPut ENDP

; ---------------------------------------------------------------------------
; DQBget SUB
; purpose:
;   Gets data for a sprite and puts the info into the given sprite array to
;   be used with DQBput. DQBget sprite format is different from the one used
;   by normal GET and PUT, so pay attention. It must be x1<x2 and y1<y2; the
;   routine is optimized if the sprite width is a multiple of 2 or 4 (better)
; declaration:
;   DECLARE SUB DQBget(BYVAL Layer,BYVAL x1,BYVAL y1,BYVAL x2,BYVAL y2,
;                       BYVAL BufferSeg,BYVAL BufferOff)
; ---------------------------------------------------------------------------
PUBLIC DQBget
DQBget PROC
  ; Stack layout (after startup):
  ;
  ; 20  (24)  Layer
  ; 18  (22)  x1
  ; 16  (20)  y1
  ; 14  (18)  x2
  ; 12  (16)  y2
  ; 10  (14)  Buffer segment
  ; 08  (12)  Buffer offset
  ; 06  (10)  Basic return segment
  ; 04  (08)  Basic return offset
  ; 02  (06)  DS
  ; 00  (04)  BP
  ;     (02)  Height
  ;     (00)  Width
  PUSH DS
  PUSH BP
  MOV BP,SP
  MOV AX,[BP+14]
  SUB AX,[BP+18]
  INC AX            ; AX holds the width
  MOV BX,[BP+12]
  SUB BX,[BP+16]
  INC BX            ; BX holds the height
  PUSH BX           ; Pushes the height onto the stack
  PUSH AX           ; Pushes the width onto the stack
  MOV BP,SP         ; Gets the new stack pointer
  MOV BX,[BP+24]
  CALL GetLayerSeg
  MOV DS,CX         ; Sets the source segment as the specified layer segment
  MOV SI,[BP+20]
  MOV BX,SI
  SHL SI,8
  SHL BX,6
  ADD SI,BX
  ADD SI,[BP+22]    ; SI now points to (x1,y1) onto the layer
  MOV BX,[BP+14]
  MOV ES,BX
  MOV DI,[BP+12]    ; ES:DI points to the buffer
  MOV DX,[BP+2]     ; DX holds the height of the sprite...
  MOV BX,[BP]       ; ...and BX its width
  SHL BX,3          ; Multiply width by 8 for GET/PUT compatibility
  MOV ES:[DI],BX    ; Stores the width and the height into the sprite buffer
  MOV ES:[DI+2],DX  ; for later use with DQBput
  ADD DI,4
  MOV BX,[BP]
  AND BX,3
  CMP BX,0
  JE DWordGetLoop   ; Width is multiple of 4: DWORDs can be used (really fast)
  MOV BX,[BP]
  AND BX,1
  CMP BX,0
  JE WordGetLoop    ; Width is multiple of 2: let's use WORDs (fast)
ByteGetLoop:
  MOV CX,[BP]       ; No optimization is possible: using bytes copy (slow)
  REP MOVSB         ; Copies "width" bytes
  SUB SI,[BP]
  ADD SI,320
  DEC DX
  JNZ ByteGetLoop
  JMP EndGet
WordGetLoop:
  MOV CX,[BP]
  SHR CX,1
  REP MOVSW         ; Copies "width\2" words
  SUB SI,[BP]
  ADD SI,320
  DEC DX
  JNZ WordGetLoop
  JMP EndGet
DWordGetLoop:
  MOV CX,[BP]
  SHR CX,2
  REP MOVSD         ; Copies "width\4" dwords
  SUB SI,[BP]
  ADD SI,320
  DEC DX
  JNZ DWordGetLoop
EndGet:
  POP BX
  POP BX
  POP BP
  POP DS
  RET 14
DQBget ENDP

; ---------------------------------------------------------------------------
; DQBsetClipBox SUB
; purpose:
;   Sets the actual clipping box used by the graphical functions.
; declaration:
;   DECLARE SUB DQBsetClipBox(BYVAL x1,BYVAL y1,BYVAL x2,BYVAL y2)
; ---------------------------------------------------------------------------
PUBLIC DQBsetClipBox
DQBsetClipBox PROC
  ; Stack layout
  ;
  ; 12  x1
  ; 10  y1
  ; 08  x2
  ; 06  y2
  ; 04  Basic return segment
  ; 02  Basic return offset
  ; 00  BP
  PUSH BP
  MOV BP,SP
  MOV AX,[BP+12]
  MOV BX,[BP+8]
  CMP BX,AX         ; Is x1>x2?
  JG CheckOther
  XCHG AX,BX        ; Yes: exchange them
CheckOther:
  MOV ClipX1,AX
  MOV ClipX2,BX
  MOV AX,[BP+10]
  MOV BX,[BP+6]
  CMP BX,AX         ; Is y1>y1?
  JG EndSetClipBox
  XCHG AX,BX        ; Yes: exchange them
EndSetClipBox:
  MOV ClipY1,AX
  MOV ClipY2,BX
  POP BP
  RET 8
DQBsetClipBox ENDP

; ---------------------------------------------------------------------------
; DQBput SUB
; purpose:
;   Draws a given sprite at specified coordinates on the given layer. If the
;   PutMode internal variable is 0, black pixels (color 0 assumed) are
;   skipped, allowing a transparent sprite drawing. The sprite automatically
;   clips when it reaches the actual clipping box borders; any x and y value
;   are accepted, as long as they're INTEGERs. The sprite can have been
;   stored into an array with DQBget as well as with GET, as the data format
;   is the same.
; declaration:
;   DECLARE SUB DQBput(BYVAL Layer,BYVAL x,BYVAL y,BYVAL BufferSeg,
;                       BYVAL BufferOff)
; ---------------------------------------------------------------------------
PUBLIC DQBput
DQBput PROC
  ; Stack layout
  ;
  ; 16  Layer
  ; 14  x
  ; 12  y
  ; 10  Buffer segment
  ; 08  Buffer offset
  ; 06  Basic return segment
  ; 04  Basic return offset
  ; 02  DS
  ; 00  BP
  PUSH DS
  PUSH BP
  MOV BP,SP
  MOV SI,[BP+08]    ; SI points to the buffer offset
  MOV BX,[BP+16]    ; Gets the layer number...
  CALL GetLayerSeg
  MOV ES,CX         ; ...and sets ES to its segment
  MOV DI,[BP+12]    ; DI contains the y value
  DEC DI
  MOV AX,PutMode    ; Puts the put mode and the clipping box coordinates
  PUSH AX           ; on the stack for later use
  MOV AX,ClipX1
  PUSH AX
  MOV AX,ClipX2
  PUSH AX
  MOV AX,ClipY1
  PUSH AX
  MOV AX,ClipY2
  PUSH AX
  MOV BX,[BP+10]
  MOV DS,BX         ; DS points to the buffer segment
  MOV BX,DI
  SHL DI,8
  SHL BX,6
  ADD DI,BX
  ADD DI,[BP+14]    ; DI points to (x,y-1) on specified layer
  LODSW
  SHR AX,3          ; Divide width by 8 for GET/PUT compatibility
  PUSH AX           ; Gets the sprite width and puts it onto the stack
  LODSW
  MOV AH,AL         ; Sets AH to the height of the sprite
  INC AH
  XOR AL,AL
  MOV BP,SP
  MOV BX,[BP+24]    ; BX holds the value y-1
  DEC BX
  ADD DI,[BP]
  CMP [BP+10],AL    ; Are we in solid put mode?
  JNE SolidYloop    ; Yes: jump to the solid put loop
Yloop:
  ADD SI,[BP]
  ADD DI,320        ; Adds 320 to DI
  INC BX            ; Next line
  DEC AH
  JZ EndPut         ; Last sprite line reached
  CMP BX,[BP+4]     ; Are we out of the upper clipping border?
  JL Yloop          ; Yes: increase the line counter
  CMP BX,[BP+2]     ; Are we out of the lower clipping border?
  JG EndPut         ; Yes: exit DQBput
  SUB DI,[BP]       ; Subtracts the sprite width to DI and SI
  SUB SI,[BP]
  MOV DX,[BP+26]    ; DX holds the current x value
  MOV CX,[BP]       ; Sets CX to the sprite width
Xloop:
  LODSB             ; Gets a pixel from the buffer
  CMP AL,0
  JE SkipPixel      ; Transparent pixel: skip it!
  CMP DX,[BP+8]
  JL SkipPixel      ; Out of left layer border: skip it!
  CMP DX,[BP+6]
  JG SkipPixel      ; Out of right layer border: skip it!
  STOSB             ; All checks done: writes pixel on the layer
  INC DX            ; Next pixel
  LOOP Xloop        ; Repeats sprite width times
  JMP Yloop         ; Next line
SkipPixel:
  INC DI            ; Increase DI without writing pixel
  INC DX            ; Next pixel
  LOOP Xloop        ; Repeats sprite width times
  JMP Yloop         ; Next line
SolidYloop:
  ADD SI,[BP]
  ADD DI,320        ; Adds 320 to DI
  INC BX            ; Next line
  DEC AH
  JZ EndPut         ; Last sprite line reached?
  CMP BX,[BP+4]     ; Are we out of the upper clipping border?
  JL SolidYloop     ; Yes: increase the line counter
  CMP BX,[BP+2]     ; Are we out of the lower clipping border?
  JG EndPut         ; Yes: exit DQBput
  SUB DI,[BP]       ; Subtracts the sprite width to DI and SI
  SUB SI,[BP]
  MOV DX,[BP+26]    ; DX holds the current x value
  MOV CX,[BP]       ; Sets CX to the sprite width
SolidXloop:
  LODSB             ; Gets a pixel from the buffer
  CMP DX,[BP+8]
  JL SkipSolidPixel    ; Out of left layer border: skip it!
  CMP DX,[BP+6]
  JG SkipSolidPixel    ; Out of right layer border: skip it!
  STOSB             ; All checks done: writes pixel on the layer
  INC DX            ; Next pixel
  LOOP SolidXloop   ; Repeats sprite width times
  JMP SolidYloop    ; Next line
SkipSolidPixel:
  INC DI            ; Increase DI without writing pixel
  INC DX            ; Next pixel
  LOOP SolidXloop   ; Repeats sprite width times
  JMP SolidYloop    ; Next line
EndPut:
  POP BX            ; Release stack memory
  POP BX
  POP BX
  POP BX
  POP BX
  POP BX
  POP BP
  POP DS
  RET 10
DQBput ENDP

; ---------------------------------------------------------------------------
; DQBxPut SUB
; purpose:
;   This function gets a sprite from a layer and puts it into another one,
;   without the support of external QB arrays. The bad thing is that to use
;   DQBxPut your sprites must not have a height bigger than 50 pixels. This
;   function also supports clipping and transparency.
; declaration:
;   DECLARE SUB DQBput(BYVAL SourceLayer,BYVAL x1,BYVAL y1,BYVAL x2,BYVAL y2,
;                      BYVAL DestLayer,BYVAL x,BYVAL y)
; ---------------------------------------------------------------------------
PUBLIC DQBxPut
DQBxPut PROC
  ; Stack layout
  ;
  ; 38  SourceLayer
  ; 36  x1
  ; 34  y1
  ; 32  x2
  ; 30  y2
  ; 28  DestLayer
  ; 26  x
  ; 24  y
  ; 22  Basic return segment
  ; 20  Basic return offset
  ; 18  DS
  ; 16  BP
  ; 14  PutMode
  ; 12  ClipX1
  ; 10  ClipX2
  ; 08  ClipY1
  ; 06  ClipY2
  ; 04  EMSseg
  ; 02  EMShdl
  ; 00  Sprite width
  PUSH DS
  PUSH BP
  MOV AX,PutMode    ; Puts the put mode and the clipping box coordinates
  PUSH AX           ; on the stack for later use
  MOV AX,ClipX1
  PUSH AX
  MOV AX,ClipX2
  PUSH AX
  MOV AX,ClipY1
  PUSH AX
  MOV AX,ClipY2
  PUSH AX
  MOV AX,EMSseg
  PUSH AX
  MOV AX,EMShdl
  PUSH AX
  PUSH AX
  MOV BP,SP
  MOV SI,[BP+34]
  MOV AX,SI
  SHL SI,8
  SHL AX,6
  ADD SI,AX
  ADD SI,[BP+36]
  MOV BX,[BP+38]
  CMP BX,[BP+28]
  JE EndXput
  CMP BX,0
  JNE sNotVideo
  MOV AX,0A000h
  MOV DS,AX
  JMP SourceFixed
sNotVideo:
  MOV AX,[BP+4]
  MOV DS,AX
  MOV AX,SI
  SHR AX,14
  CMP AX,3
  JL PageOk1
  MOV AX,2
PageOk1:
  DEC BX
  SHL BX,2
  ADD BX,AX       ; Now BX holds the correct starting page
  SHL AX,14
  SUB SI,AX       ; SI is now fixed
  MOV DX,[BP+2]
  MOV AX,04400h
  INT 67h
  INC BX
  MOV AX,04401h
  INT 67h
SourceFixed:
  MOV DI,[BP+24]
  MOV AX,DI
  SHL DI,8
  SHL AX,6
  ADD DI,AX
  ADD DI,[BP+26]
  MOV BX,[BP+28]
  CMP BX,0
  JNE dNotVideo
  MOV AX,0A000h
  MOV ES,AX
  JMP DestFixed
dNotVideo:
  MOV AX,[BP+4]
  MOV ES,AX
  MOV AX,DI
  SHR AX,14
  CMP AX,3
  JL PageOk2
  MOV AX,2
PageOk2:
  DEC BX
  SHL BX,2
  ADD BX,AX
  SHL AX,14
  SUB DI,AX
  ADD DI,08000h
  MOV DX,[BP+2]
  MOV AX,04402h
  INT 67h
  INC BX
  MOV AX,04403h
  INT 67h
DestFixed:
  MOV AX,[BP+32]
  SUB AX,[BP+36]
  INC AX
  MOV [BP],AX
  MOV AX,[BP+30]
  SUB AX,[BP+34]
  MOV AH,AL         ; Sets AH to the height of the sprite
  ADD AH,2
  XOR AL,AL
  MOV BX,[BP+24]    ; BX holds the value y-1
  DEC BX
  SUB DI,320
  SUB SI,320
  ADD DI,[BP]
  ADD SI,[BP]
  CMP [BP+14],AL    ; Are we in solid put mode?
  JNE xSolidYloop    ; Yes: jump to the solid put loop
xYloop:
  ADD SI,320
  ADD DI,320        ; Adds 320 to DI
  INC BX            ; Next line
  DEC AH
  JZ EndXput        ; Last sprite line reached
  CMP BX,[BP+8]     ; Are we out of the upper clipping border?
  JL xYloop         ; Yes: increase the line counter
  CMP BX,[BP+6]     ; Are we out of the lower clipping border?
  JG EndXput        ; Yes: exit DQBxPut
  SUB DI,[BP]       ; Subtracts the sprite width to DI and SI
  SUB SI,[BP]
  MOV DX,[BP+26]    ; DX holds the current x value
  MOV CX,[BP]       ; Sets CX to the sprite width
xXloop:
  LODSB             ; Gets a pixel from the buffer
  CMP AL,0
  JE xSkipPixel     ; Transparent pixel: skip it!
  CMP DX,[BP+12]
  JL xSkipPixel     ; Out of left layer border: skip it!
  CMP DX,[BP+10]
  JG xSkipPixel     ; Out of right layer border: skip it!
  STOSB             ; All checks done: writes pixel on the layer
  INC DX            ; Next pixel
  LOOP xXloop       ; Repeats sprite width times
  JMP xYloop        ; Next line
xSkipPixel:
  INC DI            ; Increase DI without writing pixel
  INC DX            ; Next pixel
  LOOP xXloop       ; Repeats sprite width times
  JMP xYloop        ; Next line
xSolidYloop:
  ADD SI,320
  ADD DI,320        ; Adds 320 to DI
  INC BX            ; Next line
  DEC AH
  JZ EndXput        ; Last sprite line reached?
  CMP BX,[BP+4]     ; Are we out of the upper clipping border?
  JL xSolidYloop    ; Yes: increase the line counter
  CMP BX,[BP+2]     ; Are we out of the lower clipping border?
  JG EndXput        ; Yes: exit DQBxPut
  SUB DI,[BP]       ; Subtracts the sprite width to DI and SI
  SUB SI,[BP]
  MOV DX,[BP+26]    ; DX holds the current x value
  MOV CX,[BP]       ; Sets CX to the sprite width
xSolidXloop:
  LODSB             ; Gets a pixel from the buffer
  CMP DX,[BP+12]
  JL xSkipSolidPixel   ; Out of left layer border: skip it!
  CMP DX,[BP+10]
  JG xSkipSolidPixel   ; Out of right layer border: skip it!
  STOSB             ; All checks done: writes pixel on the layer
  INC DX            ; Next pixel
  LOOP xSolidXloop  ; Repeats sprite width times
  JMP xSolidYloop   ; Next line
xSkipSolidPixel:
  INC DI            ; Increase DI without writing pixel
  INC DX            ; Next pixel
  LOOP xSolidXloop  ; Repeats sprite width times
  JMP xSolidYloop   ; Next line
EndXput:
  POP BX            ; Release stack memory
  POP BX
  POP BX
  POP BX
  POP BX
  POP BX
  POP BX
  POP BX
  POP BP
  POP DS
  MOV EMSpage,-1
  RET 16
DQBxPut ENDP

; ---------------------------------------------------------------------------
; DQBsetCol SUB
; purpose:
;   Sets the palette for a specified color
; declaration:
;   DECLARE SUB DQBsetCol(BYVAL ColorIndex,BYVAL r,BYVAL g,BYVAL b)
; ---------------------------------------------------------------------------
PUBLIC DQBsetCol
DQBsetCol PROC
  ; Stack layout:
  ;
  ; 12  Color index
  ; 10  r
  ; 08  g
  ; 06  b
  ; 04  Basic return segment
  ; 02  Basic return offset
  ; 00  BP
  PUSH BP
  MOV BP,SP
  MOV DX,3C8h       ; Port 3C8h: write palette
  MOV AX,[BP+12]
  OUT DX,AL         ; Sets the color
  MOV DX,3C9h
  MOV AL,[BP+10]
  OUT DX,AL         ; Outs the red value
  MOV AL,[BP+8]
  OUT DX,AL         ; Outs the green value
  MOV AL,[BP+6]
  OUT DX,AL         ; Outs the blue value
  POP BP
  RET 8
DQBsetCol ENDP

; ---------------------------------------------------------------------------
; DQBgetCol SUB
; purpose:
;   Gets the hues of a given color index
; declaration:
;   DECLARE SUB DQBgetCol(BYVAL ColorIndex,r,g,b)
; ---------------------------------------------------------------------------
PUBLIC DQBgetCol
DQBgetCol PROC
  ; Stack layout:
  ;
  ; 12  Color index
  ; 10  r (address)
  ; 08  g (address)
  ; 06  b (address)
  ; 04  Basic return segment
  ; 02  Basic return offset
  ; 00  BP
  PUSH BP
  MOV BP,SP
  MOV DX,3C7h       ; Port 3C7h: read palette
  MOV AX,[BP+12]
  OUT DX,AL         ; Sets the color
  MOV DX,3C9h
  XOR AX,AX
  IN AL,DX          ; Gets the red value...
  MOV BX,[BP+10]
  MOV [BX],AX       ; ...and stores it into the red variable
  IN AL,DX          ; Gets the green value...
  MOV BX,[BP+8]
  MOV [BX],AX       ; ...and stores it into the green variable
  IN AL,DX          ; Gets the blue value...
  MOV BX,[BP+6]
  MOV [BX],AX       ; ...and stores it into the blue variable
  POP BP
  RET 8
DQBgetCol ENDP

; ---------------------------------------------------------------------------
; DQBsetPal SUB
; purpose:
;   Sets the entire palette to the specified one, passed as a string*768,
;   containing in order the r,g and b data for each color. A hue is a byte and
;   ranges from 0 to 63
; declaration:
;   DECLARE SUB xDQBsetPal(BYVAL PalSeg,BYVAL PalOff)
;   DECLARE SUB DQBsetPal(Pal AS STRING)
; ---------------------------------------------------------------------------
PUBLIC xDQBsetPal
xDQBsetPal PROC
  ; Stack layout:
  ;
  ; 10  PalSeg
  ; 08  PalOff
  ; 06  Basic return segment
  ; 04  Basic return offset
  ; 02  DS
  ; 00  BP
  PUSH DS
  PUSH BP
  MOV BP,SP
  MOV AX,[BP+10]
  MOV DS,AX
  XOR AX,AX
  MOV DX,03C8h      ; Port 03C8h: write palette index selector
  OUT DX,AL         ; Starts with color 0
  MOV SI,[BP+8]     ; SI holds the address of "Pal"
  MOV CX,768        ; Repeats 768 times
  MOV DX,03C9h      ; Port 03C9h: palette entry
SetPalLoop:
  LODSB
  OUT DX,AL         ; Outs the value to the VGA port
  LOOP SetPalLoop
  POP BP
  POP DS
  RET 4
xDQBsetPal ENDP

; ---------------------------------------------------------------------------
; DQBgetPal SUB
; purpose:
;   Gets the entire palette and stores it into a given string of 768 chars,
;   ready to be used with DQBsetPal
; declaration:
;   DECLARE SUB xDQBgetPal(BYVAL PalSeg,BYVAL PalOff)
;   DECLARE SUB DQBgetPal(Pal AS STRING)
; ---------------------------------------------------------------------------
PUBLIC xDQBgetPal
xDQBgetPal PROC
  ; Stack layout:
  ;
  ; 08  PalSeg
  ; 06  PalOff
  ; 04  Basic return segment
  ; 02  Basic return offset
  ; 00  BP
  PUSH BP
  MOV BP,SP
  MOV AX,[BP+8]
  MOV ES,AX
  MOV DX,03C7h      ; Port 03C7h: read palette index selector
  XOR AX,AX
  OUT DX,AL         ; Starts with color 0
  MOV DI,[BP+6]     ; SI holds the address of "Pal"
  MOV CX,768        ; Repeats 768 times
  MOV DX,03C9h      ; Port 03C9h: palette entry
GetPalLoop:
  IN AL,DX          ; Gets a byte from the VGA port...
  STOSB             ; ...and writes it into the buffer
  LOOP GetPalLoop
  POP BP
  RET 4
xDQBgetPal ENDP

; ---------------------------------------------------------------------------
; DQBfadeIn SUB
; purpose:
;   Fades the current palette to a specified one
; declaration:
;   DECLARE SUB xDQBfadeIn(BYVAL PalSeg,BYVAL PalOff)
;   DECLARE SUB DQBfadeIn(Pal AS STRING)
; ---------------------------------------------------------------------------
PUBLIC xDQBfadeIn
xDQBfadeIn PROC
  ; Stack layout:
  ;
  ; 08  PalSeg
  ; 06  PalOff
  ; 04  Basic return segment
  ; 02  Basic return offset
  ; 00  BP
  PUSH BP
  MOV BP,SP
  MOV AX,[BP+8]
  MOV ES,AX
CheckAllCols:
  MOV DX,3DAh
WaitPal0:
  IN AL,DX
  AND AL,8
  JNZ WaitPal0
WaitPal:
  IN AL,DX
  AND AL,8
  JZ WaitPal
  XOR SI,SI         ; SI is 0 when all colors are ok
  XOR CX,CX         ; CX is the color counter
  MOV DI,[BP+6]
CheckCol:
  MOV DX,03C7h
  MOV AL,CL
  OUT DX,AL         ; Read data from color
  MOV DX,03C9h
  IN AL,DX
  MOV BL,AL         ; BL holds the red hue
  IN AL,DX
  MOV BH,AL         ; BH holds the green hue
  IN AL,DX
  MOV AH,AL         ; AH holds the blue hue
  MOV DX,03C8h
  MOV AL,CL
  OUT DX,AL         ; Write data to color
  MOV DX,03C9h
  MOV AL,ES:[DI]
  INC DI
  CMP BL,AL
  JG DecRed
  CMP BL,AL
  JL IncRed
  OUT DX,AL
  JMP CheckGreen
DecRed:
  INC SI
  MOV AL,BL
  DEC AL
  OUT DX,AL
  JMP CheckGreen
IncRed:
  INC SI
  MOV AL,BL
  INC AL
  OUT DX,AL
CheckGreen:
  MOV AL,ES:[DI]
  INC DI
  CMP BH,AL
  JG DecGreen
  CMP BH,AL
  JL IncGreen
  OUT DX,AL
  JMP CheckBlue
DecGreen:
  INC SI
  MOV AL,BH
  DEC AL
  OUT DX,AL
  JMP CheckBlue
IncGreen:
  INC SI
  MOV AL,BH
  INC AL
  OUT DX,AL
CheckBlue:
  MOV AL,ES:[DI]
  INC DI
  CMP AH,AL
  JG DecBlue
  CMP AH,AL
  JL IncBlue
  OUT DX,AL
  JMP NextCol
DecBlue:
  INC SI
  MOV AL,AH
  DEC AL
  OUT DX,AL
  JMP NextCol
IncBlue:
  INC SI
  MOV AL,AH
  INC AL
  OUT DX,AL
NextCol:
  INC CX
  CMP CX,256
  JL CheckCol
  CMP SI,0
  JNE CheckAllCols
  POP BP
  RET 4
xDQBfadeIn ENDP

; ---------------------------------------------------------------------------
; DQBfadeTo SUB
; purpose:
;   Fades all the colors of current palette to a specified color
; declaration:
;   DECLARE SUB DQBfadeTo(BYVAL Red,BYVAL Green,BYVAL Blue)
; ---------------------------------------------------------------------------
PUBLIC DQBfadeTo
DQBfadeTo PROC
  PUSH BP
  MOV BP,SP
CheckAllCols1:
  MOV DX,3DAh
WaitPal1:
  IN AL,DX
  AND AL,8
  JNZ WaitPal1
WaitPal2:
  IN AL,DX
  AND AL,8
  JZ WaitPal2
  XOR CX,CX
  XOR SI,SI
CheckCol1:
  MOV DX,03C7h
  MOV AL,CL
  OUT DX,AL
  MOV DX,03C9h
  IN AL,DX
  MOV BL,AL         ; BL holds the current red hue
  IN AL,DX
  MOV BH,AL         ; BH holds the current green hue
  IN AL,DX
  MOV AH,AL         ; AH holds the current blue hue
  MOV DX,03C8h
  MOV AL,CL
  OUT DX,AL
  MOV DX,03C9h
  MOV AL,[BP+10]
  CMP BL,AL
  JG DecRed1
  CMP BL,AL
  JL IncRed1
  OUT DX,AL
  JMP CheckGreen1
DecRed1:
  INC SI
  MOV AL,BL
  DEC AL
  OUT DX,AL
  JMP CheckGreen1
IncRed1:
  INC SI
  MOV AL,BL
  INC AL
  OUT DX,AL
CheckGreen1:
  MOV AL,[BP+8]
  CMP BH,AL
  JG DecGreen1
  CMP BH,AL
  JL IncGreen1
  OUT DX,AL
  JMP CheckBlue1
DecGreen1:
  INC SI
  MOV AL,BH
  DEC AL
  OUT DX,AL
  JMP CheckBlue1
IncGreen1:
  INC SI
  MOV AL,BH
  INC AL
  OUT DX,AL
CheckBlue1:
  MOV AL,[BP+6]
  CMP AH,AL
  JG DecBlue1
  CMP AH,AL
  JL IncBlue1
  OUT DX,AL
  JMP NextCol1
DecBlue1:
  INC SI
  MOV AL,AH
  DEC AL
  OUT DX,AL
  JMP NextCol1
IncBlue1:
  INC SI
  MOV AL,AH
  INC AL
  OUT DX,AL
NextCol1:
  INC CX
  CMP CX,256
  JL CheckCol1
  CMP SI,0
  JNE CheckAllCols1
  POP BP
  RET 6
DQBfadeTo ENDP

; ---------------------------------------------------------------------------
; DQBpalOff SUB
; purpose:
;   Turns all the colors in current palette to black.
; declaration:
;   DECLARE SUB DQBpalOff()
; ---------------------------------------------------------------------------
PUBLIC DQBpalOff
DQBpalOff PROC
  MOV DX,03C8h
  XOR AX,AX
  OUT DX,AL
  MOV DX,03C9h
  MOV CX,768
PalOff:
  OUT DX,AL
  LOOP PalOff
  RET
DQBpalOff ENDP

; ---------------------------------------------------------------------------
; DQBprint SUB
; purpose:
;   Prints a string onto specified layer, at the given coordinates and with
;   the given color and current font
; declaration:
;   DECLARE SUB xDQBprint(BYVAL Layer,BYVAL TextSeg,BYVAL TextOff,BYVAL x,
;                         BYVAL y,BYVAL Col,BYVAL Style)
;   DECLARE SUB DQBprint(Layer AS INTEGER,Text AS STRING,x AS INTEGER,y AS
;                         INTEGER,Col AS INTEGER)
;   DECLARE SUB DQBprints(Layer AS INTEGER,Text AS STRING,x AS INTEGER,y AS
;                         INTEGER,Col AS INTEGER,Style AS INTEGER)
; ---------------------------------------------------------------------------
PUBLIC xDQBprint
xDQBprint PROC
  ; Stack layout:
  ;
  ; 18  Layer
  ; 16  TextSeg
  ; 14  TextOff
  ; 12  x
  ; 10  y
  ; 08  Col
  ; 06  Style
  ; 04  Basic return segment
  ; 02  Basic return offset
  ; 00  BP
  PUSH BP
  MOV BP,SP
  MOV BX,[BP+18]
  CALL GetLayerSeg
  MOV ES,CX
  MOV AX,[BP+16]
  MOV FS,AX         ; FS holds the text segment
  MOV SI,[BP+14]    ; SI holds the character offset
  MOV AX,[BP+6]
  MOV BL,TextMode
  PUSH BX
  MOV BX,[BP+12]    ; BX holds the character x position
  CMP AL,0FFh
  JE OldStyle
  MOV TextMode,AL
OldStyle:
  TEST TextMode,4
  JZ GetChar
  XOR AH,AH
  MOV AL,CharLen[256]
  SHR AX,2
  ADD BX,AX         ; Fix x position for italic text style
GetChar:
  XOR AH,AH
  MOV AL,FS:[SI]    ; Gets character
  CMP AL,0
  JE ExitPrint      ; If null character then exit
  PUSH SI
  SHL AX,3
  ADD AX,OFFSET FontBuffer
  MOV SI,AX         ; Now SI holds the character data offset
  MOV DX,[BP+10]     ; DX holds the y position
  XOR CH,CH
  MOV CL,CharLen[256]    ; Char line counter
DrawCharLine:
  CMP DX,ClipY2
  JG DrawNextChar
  PUSH CX
  CMP DX,ClipY1
  JL NextCharLine
  MOV AL,[SI]       ; AL holds the character line data
  MOV DI,DX
  MOV CX,DI
  SHL DI,8
  SHL CX,6
  ADD DI,CX
  ADD DI,BX
  PUSH BX
  PUSH DX
  XOR DX,DX
CharLineLoop:
  MOV CX,7
  SUB CX,DX
  MOV AH,1
  SHL AH,CL
  TEST AL,AH
  JZ TestBack
  CMP BX,ClipX1
  JL SkipCharPixel
  CMP BX,ClipX2
  JG SkipCharPixel
  PUSH AX
  MOV AX,[BP+8]
  MOV ES:[DI],AL
  POP AX
  TEST TextMode,2
  JNZ BoldEffect
  JMP SkipCharPixel
TestBack:
  TEST TextMode,1
  JZ SkipCharPixel
  CMP BX,ClipX1
  JL SkipCharPixel
  CMP BX,ClipX2
  JG SkipCharPixel
  PUSH AX
  MOV AL,TextBackCol
  MOV ES:[DI],AL
  POP AX
SkipCharPixel:
  INC DI
  INC BX
  INC DX
  CMP DX,8
  JL CharLineLoop
  POP DX
  POP BX
NextCharLine:
  INC DX
  INC SI
  POP CX
  TEST TextMode,4
  JZ NoItalicFix
  PUSH AX
  MOV AX,CX
  DEC AX
  AND AX,3
  CMP AX,3
  JNE NoItalicFix1
  DEC BX
NoItalicFix1:
  POP AX
NoItalicFix:
  DEC CL
  JNZ DrawCharLine
DrawNextChar:
  POP SI
  XOR AH,AH
  MOV AL,FS:[SI]
  MOV DI,AX
  MOV AL,CharLen[DI]
  XOR AH,AH
  ADD BX,AX         ; Increases x position
  TEST TextMode,8
  JZ NoUnderFix
  CMP DX,ClipY1
  JL NoUnderFix
  CMP DX,ClipY2
  JG NoUnderFix
  PUSH AX
  SUB BX,AX
  MOV CX,AX
  ADD CX,3
  MOV DI,DX
  MOV AX,DI
  SHL DI,8
  SHL AX,6
  ADD DI,AX
  ADD DI,BX
  DEC DI
  DEC BX
  MOV AL,[BP+8]
UnderLine:
  INC DI
  INC BX
  DEC CX
  JZ UnderLineDone
  CMP BX,ClipX1
  JL UnderLine
  CMP BX,ClipX2
  JG UnderLine
  MOV ES:[DI],AL
  JMP UnderLine
UnderLineDone:
  POP AX
  SUB BX,2
NoUnderFix:
  TEST TextMode,2
  JZ NoBoldFix
  INC BX
NoBoldFix:
  TEST TextMode,4
  JZ NoItalicFix2
  ADD BX,2
NoItalicFix2:
  INC SI
  JMP GetChar
ExitPrint:
  POP BX
  MOV TextMode,BL
  POP BP
  RET 14
BoldEffect:
  INC BX
  CMP BX,ClipX1
  JL SkipCharPixel
  CMP BX,ClipX2
  JG SkipCharPixel
  PUSH AX
  MOV AX,[BP+8]
  MOV ES:[DI+1],AL
  POP AX
  DEC BX
  JMP SkipCharPixel
xDQBprint ENDP

; ---------------------------------------------------------------------------
; DQBlen FUNCTION
; purpose:
;   Restores the length of a string in pixels, by checking the current font
;   characters length table.
; declaration:
;   DECLARE FUNCTION xDQBlen(BYVAL TextSeg,BYVAL TextOff)
;   DECLARE FUNCTION DQBlen(Text AS STRING)
; ---------------------------------------------------------------------------
PUBLIC xDQBlen
xDQBlen PROC
  PUSH BP
  MOV BP,SP
  MOV AX,[BP+8]
  MOV ES,AX
  MOV SI,[BP+6]
  XOR CX,CX
  XOR AH,AH
  XOR DH,DH
  CMP ES:[SI],AH
  JE EndLen
  TEST TextMode,2
  JZ NoLenFix
  INC CX
NoLenFix:
  TEST TextMode,4
  JZ FindLen
  MOV AL,CharLen[256]
  SHR AL,2
  ADD CX,AX
FindLen:
  MOV AL,ES:[SI]
  INC SI
  CMP AL,0
  JE EndLen
  MOV DI,AX
  MOV DL,CharLen[DI]
  ADD CX,DX
  JMP FindLen
EndLen:
  MOV AX,CX
  POP BP
  RET 4
xDQBlen ENDP

; ---------------------------------------------------------------------------
; DQBsetBIOSfont SUB
; purpose:
;   Restores the old VGA BIOS font as the actual font
; declaration:
;   DECLARE SUB DQBsetBIOSfont()
; ---------------------------------------------------------------------------
PUBLIC DQBsetBIOSfont
DQBsetBIOSfont PROC
  PUSH DS
  MOV AX,@DATA
  MOV ES,AX
  MOV DI,OFFSET FontBuffer
  MOV AX,0FFA6h
  MOV DS,AX
  MOV SI,0Eh
  MOV CX,512
  REP MOVSD
  MOV EAX,08080808h
  MOV DI,OFFSET CharLen
  STOSB
  MOV CX,64
  REP STOSD
  POP DS
  MOV TextMode,0
  RET
DQBsetBIOSfont ENDP

; ---------------------------------------------------------------------------
; DQBsetFont SUB
; purpose:
;   Sets the actual font to a new one stored in memory
; declaration:
;   DECLARE SUB xDQBsetFont(BYVAL FontSeg,BYVAL FontOff)
;   DECLARE SUB DQBsetFont(Font AS STRING)
; ---------------------------------------------------------------------------
PUBLIC xDQBsetFont
xDQBsetFont PROC
  ; Stack layout:
  ;
  ; 10  FontSeg
  ; 08  FontOff
  ; 06  Basic return segment
  ; 04  Basic return offset
  ; 02  DS
  ; 00  BP
  PUSH DS
  PUSH BP
  MOV BP,SP
  MOV AX,@DATA
  MOV ES,AX
  MOV DI,OFFSET FontBuffer
  MOV AX,[BP+10]
  MOV DS,AX
  MOV SI,[BP+8]
  MOV CX,512
  REP MOVSD
  MOV DI,OFFSET CharLen
  MOVSB
  MOV CX,64
  REP MOVSD
  POP BP
  POP DS
  RET 4
xDQBsetFont ENDP

; ---------------------------------------------------------------------------
; DQBsetBackTextCol SUB
; purpose:
;   Sets the color used to fill the background of text print in solid mode
; declaration:
;   DECLARE SUB DQBsetBackTextCol(BYVAL Col)
; ---------------------------------------------------------------------------
PUBLIC DQBsetTextBackCol
DQBsetTextBackCol PROC
  PUSH BP
  MOV BP,SP
  MOV AX,[BP+6]
  MOV TextBackCol,AL
  POP BP
  RET 2
DQBsetTextBackCol ENDP

; ---------------------------------------------------------------------------
; DQBsetTextStyle SUB
; purpose:
;   Sets the current text style for the DQBprint function
; declaration:
;   DECLARE SUB DQBsetTextStyle(BYVAL Style)
; ---------------------------------------------------------------------------
PUBLIC DQBsetTextStyle
DQBsetTextStyle PROC
  PUSH BP
  MOV BP,SP
  MOV AX,[BP+6]
  MOV TextMode,AL
  POP BP
  RET 2
DQBsetTextStyle ENDP

; ---------------------------------------------------------------------------
; DQBloadLayer FUNCTION
; purpose:
;   Loads an image onto specified layer. It supports BSAVE, BMP and PCX image
;   formats, and it automatically detects which one is used by specified file.
;   When loading a BMP or a PCX, its palette is stored at specified address;
;   when loading BSAVEd images, the palette is ignored.
; declaration:
;   DECLARE FUNCTION xDQBloadLayer(BYVAL Layer,BYVAL FileSeg,BYVAL FileOff,
;                                 BYVAL PalSeg,BYVAL PalOff)
;   DECLARE FUNCTION DQBloadLayer(Layer AS INTEGER,FileName AS STRING,
;                                 Pal AS STRING)
; ---------------------------------------------------------------------------
PUBLIC xDQBloadLayer
xDQBloadLayer PROC
  ; Stack layout:
  ;
  ; 16  Layer
  ; 14  FileSeg
  ; 12  FileOff
  ; 10  PalSeg
  ; 08  PalOff
  ; 06  Basic return segment
  ; 04  Basic return offset
  ; 02  DS
  ; 00  BP
  PUSH DS
  PUSH BP
  MOV BP,SP
  CALL HookInt24
  MOV BX,[BP+16]
  CALL GetLayerSeg
  MOV ES,CX
  MOV AX,[BP+14]
  MOV DS,AX
  MOV DX,[BP+12]
  MOV AX,3D00h
  INT 21h
  JC Error_Open
  MOV BX,AX
  MOV AX,@DATA
  MOV DS,AX
  XOR AX,AX
  XOR DX,DX
  XOR SI,SI
  CLD
FindFormat:
  CALL ReadBuffer
  JC Error_Read
  MOV CL,Buffer
  CMP CL,BSVheader[SI]
  JNE NotBSV
  INC AH
NotBSV:
  CMP CL,BMPheader[SI]
  JNE NotBMP
  INC DL
NotBMP:
  CMP CL,PCXheader[SI]
  JNE NotPCX
  INC AL
NotPCX:
  INC SI
  CMP SI,7
  JL FindFormat
  CMP AH,7
  JE LoadBSV
  CMP DL,7
  JE LoadBMP
  CMP AL,7
  JE LoadPCX
  JMP Error_BadFile
LoadBMP:
  XOR SI,SI
  MOV CX,47         ; Skip remaining of BMP header
CheckBMPheader:
  CALL ReadBuffer
  JC Error_Read
  LOOP CheckBMPheader
  PUSH ES
  MOV AX,[BP+10]
  MOV ES,AX
  MOV DI,[BP+8]
  MOV CX,256
LoadPalette:
  CALL ReadBuffer
  JC Error_Read
  MOV AL,Buffer
  SHR AL,2
  MOV ES:[DI+2],AL
  CALL ReadBuffer
  JC Error_Read
  MOV AL,Buffer
  SHR AL,2
  MOV ES:[DI+1],AL
  CALL ReadBuffer
  JC Error_Read
  MOV AL,Buffer
  SHR AL,2
  MOV ES:[DI],AL
  CALL ReadBuffer
  JC Error_Read
  ADD DI,3
  LOOP LoadPalette
  POP ES
  MOV CX,200
LoadBMPloop:
  MOV DX,CX
  DEC DX
  MOV SI,DX
  SHL DX,8
  SHL SI,6
  ADD DX,SI
  PUSH CX
  MOV AX,ES
  MOV DS,AX
  MOV CX,320
  MOV AH,3Fh
  INT 21h
  JC Error_Read
  POP CX
  LOOP LoadBMPloop
  JMP LoadCloseFile
LoadPCX:
  XOR SI,SI
  MOV CX,121        ; Skip remaining of PCX header
CheckPCXheader:
  CALL ReadBuffer
  JC Error_Read
  LOOP CheckPCXheader
  XOR DI,DI
  XOR AX,AX
LoadPCXloop:
  CALL ReadBuffer
  JC Error_Read
  MOV AL,Buffer
  PUSH AX
  AND AL,0C0h
  CMP AL,0C0h
  POP AX
  JE DecodeChunk
  STOSB
  JMP NextChunk
DecodeChunk:
  SUB AL,0C0h
  PUSH AX
  CALL ReadBuffer
  JC Error_Read
  POP CX
  MOV AL,Buffer
  REP STOSB
NextChunk:
  CMP DI,0FA00h
  JB LoadPCXloop
  CALL ReadBuffer
  JC Error_Read
  MOV AX,[BP+10]
  MOV ES,AX
  MOV DI,[BP+8]
  MOV CX,768
LoadPalette1:
  CALL ReadBuffer
  JC Error_Read
  MOV AL,Buffer
  SHR AL,2
  STOSB
  LOOP LoadPalette1
  JMP LoadCloseFile
LoadBSV:
  MOV AX,ES
  MOV DS,AX
  MOV CX,64000
  XOR DX,DX
  MOV AH,3Fh
  INT 21h
  JC Error_Read
  JMP LoadCloseFile
Error_Open:
  MOV AX,1          ; Error 1: Cannot open file (does it exist?)
  JMP EndLoadLayer
Error_Read:
  MOV AH,3Eh
  INT 21h
  MOV AX,2          ; Error 2: Unable to read from file
  JMP EndLoadLayer
Error_BadFile:
  MOV AH,3Eh
  INT 21h
  MOV AX,3          ; Error 3: Bad file format
  JMP EndLoadLayer
LoadCloseFile:
  MOV AH,3Eh
  INT 21h
  XOR AX,AX
EndLoadLayer:
  CALL UnhookInt24
  POP BP
  POP DS
  RET 10
xDQBloadLayer ENDP

; ---------------------------------------------------------------------------
; DQBsaveLayer FUNCTION
; purpose:
;   Saves specified layer to a file. Two saving methods available: the first
;   saves only the layer into a BSAVE compatible binary file, the second
;   stores the layer as a PCX file, including specified palette. The function
;   returns 0 on success, otherwise an error code.
; declaration:
;   DECLARE FUNCTION xDQBsaveLayer(BYVAL Layer,BYVAL FileSeg,BYVAL FileOff,
;                                 BYVAL PalSeg,BYVAL PalOff,BYVAL Format)
;   DECLARE FUNCTION DQBsaveLayer(Layer AS INTEGER,FileName AS STRING,
;                                 Pal AS STRING,Format AS INTEGER)
; ---------------------------------------------------------------------------
PUBLIC xDQBsaveLayer
xDQBsaveLayer PROC
  ; Stack layout:
  ;
  ; 18  Layer
  ; 16  FileSeg
  ; 14  FileOff
  ; 12  PalSeg
  ; 10  PalOff
  ; 08  Format
  ; 06  Basic return segment
  ; 04  Basic return offset
  ; 02  DS
  ; 00  BP
  PUSH DS
  PUSH BP
  MOV BP,SP
  CALL HookInt24
  MOV BX,[BP+18]
  CALL GetLayerSeg
  MOV ES,CX
  MOV DX,[BP+14]    ; DS:DX points to the asciiz filename
  MOV AX,[BP+16]
  MOV DS,AX
  MOV AH,3Ch        ; Function 3Ch: create file
  MOV CX,20h        ; archive attribute
  INT 21h
  JC Error_Create
  MOV BX,AX
  MOV AX,[BP+8]
  CMP AX,0
  JE SaveBSV
  CMP AX,1
  JE SaveBMP
  MOV DX,OFFSET PCXheader
  MOV AX,@DATA
  MOV DS,AX
  MOV CX,128        ; Let's write the PCX header at the beginning of the file
  MOV AH,40h        ; Function 40h: write to file
  INT 21h
  JC Error_Write
  XOR DI,DI         ; Let's begin with the first layer pixel
  MOV CX,200        ; Scans 200 lines
SavePCXloop:
  XOR DX,DX         ; DX holds the pixel being processed
EncodeLine:
  XOR SI,SI         ; SI holds the number of consecutive equal pixels
  PUSH BX
FindEquals:
  MOV BX,DX
  ADD BX,SI
  CMP BX,320        ; Have we reached the end of line?
  JGE WriteData      ; Yes: write data
  CMP SI,63         ; A maximum of 63 consecutive pixels is allowed
  JGE WriteData
  MOV BX,DI
  ADD BX,SI
  MOV AL,ES:[BX]
  CMP AL,ES:[BX+1]  ; Let's check if we have two equal consecutive pixels
  JNE WriteData     ; No: write a single pixel
  INC SI            ; Yes: increase the consecutive pixels counter (SI)
  JMP FindEquals
WriteData:
  CMP SI,0          ; Are there any consecutive equal pixels?
  JE SinglePixel    ; No: write a single pixel
  ADD DX,SI
  ADD DI,SI
  MOV BX,SI
  OR BL,192
  MOV AH,BL
  POP BX
  MOV Buffer,AH
  CALL WriteBuffer
  JC Error_Write
  MOV Buffer,AL
  CALL WriteBuffer
  JC Error_Write
  JMP EncodeNext
SinglePixel:
  POP BX
  INC DX
  INC DI
  TEST AL,192
  JZ SinglePixel1
  MOV Buffer,193
  CALL WriteBuffer
  JC Error_Write
  JMP SinglePixel1
EncodeNext:
  CMP DX,320
  JL EncodeLine
  LOOP SavePCXloop
  JMP SavePalette
SinglePixel1:
  MOV Buffer,AL
  CALL WriteBuffer
  JC Error_Write
  JMP EncodeNext
SavePalette:
  MOV Buffer,12
  CALL WriteBuffer
  JC Error_Write
  MOV DX,[BP+12]
  MOV DS,DX
  MOV DX,@DATA
  MOV SI,[BP+10]
  MOV CX,768
WritePalette:
  LODSB
  SHL AL,2
  PUSH DS
  MOV DS,DX
  MOV Buffer,AL
  CALL WriteBuffer
  JC Error_Write
  POP DS
  DEC CX
  JNZ WritePalette
  JMP SaveCloseFile
SaveBMP:
  MOV CX,54
  MOV DX,OFFSET BMPheader
  MOV AX,@DATA
  MOV DS,AX
  MOV AH,40h        ; Saves the BMP header
  INT 21h
  JC Error_Write
  MOV SI,[BP+10]    ; DS:SI now points to the palette string
  MOV CX,256
WritePalette1:
  MOV AX,[BP+12]
  MOV DS,AX
  LODSB
  PUSH AX
  LODSB
  PUSH AX
  LODSB
  MOV DX,@DATA
  MOV DS,DX
  SHL AL,2
  MOV Buffer,AL
  CALL WriteBuffer
  JC Error_Write
  POP AX
  SHL AL,2
  MOV Buffer,AL
  CALL WriteBuffer
  JC Error_Write
  POP AX
  SHL AL,2
  MOV Buffer,AL
  CALL WriteBuffer
  JC Error_Write
  MOV Buffer,0
  CALL WriteBuffer
  JC Error_Write
  LOOP WritePalette1
  MOV CX,200
SaveBMPloop:
  MOV DX,CX
  DEC DX
  MOV SI,DX
  SHL DX,8
  SHL SI,6
  ADD DX,SI
  PUSH CX
  MOV AX,ES
  MOV DS,AX
  MOV AH,40h
  MOV CX,320
  INT 21h
  JC Error_Write
  POP CX
  LOOP SaveBMPloop
  JMP SaveCloseFile
SaveBSV:
  MOV CX,7
  MOV DX,OFFSET BSVheader
  MOV AX,@DATA
  MOV DS,AX
  MOV AH,40h        ; Saves the BSAVE header
  INT 21h
  JC Error_Write
  MOV CX,64000
  XOR DX,DX
  MOV AX,ES
  MOV DS,AX
  MOV AH,40h        ; Saves the entire layer
  INT 21h
  JC Error_Write
  JMP SaveCloseFile
Error_Create:
  MOV AX,1          ; Error 1: cannot create file
  JMP EndSaveLayer
Error_Write:
  MOV AH,3Eh
  INT 21h
  MOV AX,2          ; Error 2: cannot write to disk (disk full?)
  JMP EndSaveLayer
SaveCloseFile:
  MOV AH,3Eh        ; Function 3Eh: close file
  INT 21h
  XOR AX,AX
EndSaveLayer:
  CALL UnhookInt24
  POP BP
  POP DS
  RET 12
xDQBsaveLayer ENDP


; ***************************************************************************
; INPUT ROUTINES
; ***************************************************************************


; ---------------------------------------------------------------------------
; DQBinstallKeyboard SUB
; purpose:
;   Installs a keyboard interrupt handler that updates the status of each key
; declaration:
;   DECLARE SUB DQBinstallKeyboard()
; ---------------------------------------------------------------------------
PUBLIC DQBinstallKeyboard
DQBinstallKeyboard PROC
  CMP KeyActive,1
  JE EndKeyInstall
  MOV AX,040h
  MOV GS,AX
  MOV SI,017h
  MOV AL,GS:[SI]
  AND AL,070h
  MOV KeyFlags,AL 
  PUSH DS
  MOV AX,3509h
  INT 21h
  POP DS
  MOV OldInt9seg,ES
  MOV OldInt9off,BX
  MOV AX,SEG KeyboardISR
  MOV DX,OFFSET KeyboardISR
  PUSH DS
  MOV DS,AX
  MOV AX,2509h
  INT 21h
  POP DS
  MOV KeyActive,1
  RET
KeyboardISR:
  PUSHF
  PUSHA
  PUSH DS
  MOV BL,1
  IN AL,60h
  TEST AL,80h
  JZ StoreStatus
  AND AL,7Fh
  MOV BL,0
StoreStatus:
  XOR AH,AH
  MOV SI,AX
  MOV CX,@DATA
  MOV DS,CX
  MOV KeyStatus[SI],BL
  IN AL,61h
  OR AL,82h
  OUT 61h,AL
  AND AL,7Fh
  OUT 61h,AL
  MOV AL,20h
  OUT 20h,AL
  POP DS
  POPA
  POPF
  IRET
EndKeyInstall:
  RET
DQBinstallKeyboard ENDP

; ---------------------------------------------------------------------------
; DQBremoveKeyboard SUB
; purpose:
;   Removes previously installed keyboard interrupt handler and restores
;   the old one
; declaration:
;   DECLARE SUB DQBremoveKeyboard()
; ---------------------------------------------------------------------------
PUBLIC DQBremoveKeyboard
DQBremoveKeyboard PROC
  CMP KeyActive,0
  JE EndKeyRemove
  PUSH DS
  MOV DX,OldInt9off
  MOV AX,OldInt9seg
  MOV DS,AX
  MOV AX,2509h
  INT 21h
  POP DS
  MOV AX,040h
  MOV ES,AX
  MOV SI,017h
  MOV AL,KeyFlags
  MOV ES:[SI],AL
  XOR SI,SI
EmptyKeyStatus:
  MOV KeyStatus[SI],0
  INC SI
  CMP SI,128
  JL EmptyKeyStatus
  MOV KeyActive,0
EndKeyRemove:
  RET
DQBremoveKeyboard ENDP

; ---------------------------------------------------------------------------
; DQBkey FUNCTION
; purpose:
;   Returns the current status of a given key
; declaration:
;   DECLARE FUNCTION DQBkey(BYVAL ScanCode)
; ---------------------------------------------------------------------------
PUBLIC DQBkey
DQBkey PROC
  ; Stack layout:
  ;
  ; 06  ScanCode
  ; 04  Basic return segment
  ; 02  Basic return offset
  ; 00 BP
  PUSH BP
  MOV BP,SP
  XOR AH,AH
  MOV SI,[BP+6]
  CMP KeyStatus[SI],1
  SETE AL
  NEG AX
  POP BP
  RET 2
DQBkey ENDP


; ***************************************************************************
; SOUND ROUTINES
; ***************************************************************************


; ---------------------------------------------------------------------------
; DQBinstallSB FUNCTION
; purpose:
;   Initialized the SB and starts the realtime mixing. Returns 0 on successful
;   otherwise an error code.
; declaration:
;   DECLARE FUNCTION DQBinstallSB(BYVAL Channels,BYVAL BaseAddr,BYVAL IRQ,
;                                 BYVAL DMA)
; ---------------------------------------------------------------------------
PUBLIC xDQBinstallSB
xDQBinstallSB PROC
  ; Stack layout:
  ;
  ; 12  Channels
  ; 10  BaseAddr
  ; 08  IRQ
  ; 06  DMA
  ; 04  Basic return segment
  ; 02  Basic return offset
  ; 00  BP
  PUSH BP
  MOV BP,SP
  CMP EMSallocated,0
  JG CheckSounds
  MOV AX,1          ; Error 1: No sounds were initialized
  POP BP
  RET 8
CheckSounds:
  CMP NumSounds,0
  JG InstallSB
  MOV AX,1          ; Error 1 again
  POP BP
  RET 8
InstallSB:
  CMP SoundActive,0
  JE NotInstalled
  XOR AX,AX
  POP BP
  RET 8
NotInstalled:
  MOV Playing,0
  MOV AX,[BP+10]
  MOV BaseAddr,AX
  MOV CX,[BP+8]
  MOV IRQ,CL
  MOV BL,1
  SHL BL,CL
  MOV PicORmask,BL
  XOR BL,0FFh
  MOV PicANDmask,BL
  MOV AX,[BP+6]
  MOV DMA,AL
  CMP AL,0
  JNE TestDMA1
  MOV DMA_Page,087h
  MOV DMA_Base,0
  MOV DMA_Count,1
  JMP DMAdone
TestDMA1:
  CMP AL,1
  JNE TestDMA2
  MOV DMA_Page,083h
  MOV DMA_Base,2
  MOV DMA_Count,3
  JMP DMAdone
TestDMA2:
  CMP AL,2
  JNE TestDMA3
  MOV DMA_Page,081h
  MOV DMA_Base,4
  MOV DMA_Count,5
  JMP DMAdone
TestDMA3:
  CMP AL,3
  JNE DMAerror
  MOV DMA_Page,082h
  MOV DMA_Base,6
  MOV DMA_Count,7
  JMP DMAdone
DMAerror:
  MOV AX,4          ; Error 4: only DMA channels 0-3 are supported
  POP BP
  RET 8
DMAdone:
  MOV DX,BaseAddr
  ADD DX,06h
  MOV AL,1
  OUT DX,AL
  IN AL,DX
  IN AL,DX
  IN AL,DX
  IN AL,DX
  XOR AL,AL
  OUT DX,AL
  ADD DX,08h
  WAITREADDSP
  SUB DX,4
  IN AL,DX
  CMP AL,0AAh       ; Has the DSP reset successfully?
  JE resetOk
  MOV AX,2          ; Error 2: DSP failed to reset
  POP BP
  RET 8
resetOk:            ; All ok: installs our new IRQ
  MOV DX,[BP+12]
  MOV Channels,DX
  MOV DX,BaseAddr
  ADD DX,0Ch
  WAITWRITEDSP
  MOV AL,0E1h       ; Get DSP version
  OUT DX,AL
  ADD DX,2
  WAITREADDSP
  SUB DX,4
  IN AL,DX
  MOV BH,AL
  ADD DX,4
  WAITREADDSP
  SUB DX,4
  IN AL,DX
  MOV BL,AL
  CMP BX,0200h
  JGE DSPok
  MOV AX,3          ; Error 3: Old soundblaster versions not supported
  POP BP
  RET 8
DSPok:
  ADD DX,2
  WAITWRITEDSP
  MOV AL,0D1h
  OUT DX,AL         ; Turns speakers on
  XOR AX,AX
  MOV ES,AX
  MOV AL,IRQ
  ADD AL,8
  MOV SI,AX
  SHL SI,2
  IN AL,021h
  AND AL,PicANDmask
  OUT 021h,AL       ; Disable IRQ
  MOV AX,ES:[SI]
  MOV OldIRQoff,AX
  MOV AX,OFFSET SBmix
  CLI
  MOV ES:[SI],AX
  MOV AX,ES:[SI+2]
  MOV OldIRQseg,AX
  MOV AX,SEG SBmix
  MOV ES:[SI+2],AX  ; We've changed the IRQ vector
  STI
  IN AL,021h
  AND AL,PicANDmask
  OUT 021h,AL       ; Enable IRQ
  SETUPDMA          ; Starts DMA transfers
  MOV SoundActive,1
  XOR AX,AX
  POP BP
  RET 8
SBmix:
  PUSHF
  PUSHA
  PUSH DS
  PUSH ES
  MOV AX,@DATA
  MOV DS,AX
  MOV ES,AX
  MOV DX,BaseAddr
  ADD DX,0Eh        ; Acknowledge SB
  IN AL,DX
  MOV SI,OFFSET SoundBuffer2      ; Double buffering to avoid "clicks"
  MOV DI,OFFSET SoundBuffer
  MOV CX,64
  REP MOVSD
  MOV AL,20h        ; Acknowledge interrupt
  OUT 20h,AL
  MOV AH,47h
  MOV DX,EMShdl
  INT 67h
  MIXVOICES         ; Mix into the second buffer while the first is playing
  MOV AH,48h
  MOV DX,EMShdl
  INT 67h
  POP ES
  POP DS
  POPA
  POPF
  IRET
xDQBinstallSB ENDP

; ---------------------------------------------------------------------------
; DQBsetSamplingRate SUB
; purpose:
;   Sets the sound sampling rate
; declaration:
;   DECLARE SUB xDQBsetSamplingRate(BYVAL Rate)
;   DECLARE SUB DQBsetSamplingRate(Rate)
; ---------------------------------------------------------------------------
PUBLIC xDQBsetSamplingRate
xDQBsetSamplingRate PROC
  PUSH BP
  MOV BP,SP
  MOV DX,BaseAddr
  ADD DX,0Ch
  WAITWRITEDSP
  MOV AL,0D0h
  OUT DX,AL
  WAITWRITEDSP
  MOV AL,040h
  OUT DX,AL
  WAITWRITEDSP
  MOV AX,[BP+6]
  OUT DX,AL
  MOV SamplingRate,AL
  WAITWRITEDSP
  MOV AL,0D4h
  OUT DX,AL
  POP BP
  RET 2
xDQBsetSamplingRate ENDP

; ---------------------------------------------------------------------------
; DQBloadSound FUNCTION
; purpose:
;   Loads a sound sample into a specified sound slot. Sounds must be 8 bit
;   mono and their sampling rate must not be greater than 22000 Hz.
; declaration:
;   DECLARE FUNCTION xDQBloadSound(BYVAL Slot,BYVAL FileSeg,BYVAL FileOff)
;   DECLARE FUNCTION DQBloadSound(Slot AS INTEGER,FileName AS STRING)
; ---------------------------------------------------------------------------
PUBLIC xDQBloadSound
xDQBloadSound PROC
  ; Stack layout
  ;
  ; 12  Slot
  ; 10  FileSeg
  ; 08  FileOff
  ; 06  Basic return segment
  ; 04  Basic return offset
  ; 02  DS
  ; 00  BP
  PUSH DS
  PUSH BP
  MOV BP,SP
  CALL HookInt24
  MOV AX,[BP+12]
  XOR BH,BH
  MOV BL,NumSounds
  CMP AX,BX
  JG Error_BadSlot
  MOV AX,[BP+10]
  MOV DS,AX
  MOV DX,[BP+8]
  MOV AX,03D00h
  INT 21h
  JC Error_OpenSound
  MOV BX,AX
  MOV AX,@DATA
  MOV DS,AX
  CALL ReadBuffer
  JC Error_ReadSound
  CMP Buffer,'R'
  JNE Error_BadSound
  CALL ReadBuffer
  JC Error_ReadSound
  CMP Buffer,'I'
  JNE Error_BadSound
  CALL ReadBuffer
  JC Error_ReadSound
  CMP Buffer,'F'
  JNE Error_BadSound
  CALL ReadBuffer
  JC Error_ReadSound
  CMP Buffer,'F'
  JNE Error_BadSound
  MOV CX,4
Skiphd1:
  CALL ReadBuffer
  JC Error_ReadSound
  LOOP Skiphd1
  CALL ReadBuffer
  JC Error_ReadSound
  CMP Buffer,'W'
  JNE Error_BadSound
  CALL ReadBuffer
  JC Error_ReadSound
  CMP Buffer,'A'
  JNE Error_BadSound
  CALL ReadBuffer
  JC Error_ReadSound
  CMP Buffer,'V'
  JNE Error_BadSound
  CALL ReadBuffer
  JC Error_ReadSound
  CMP Buffer,'E'
  JNE Error_BadSound
  MOV CX,10
Skiphd2:
  CALL ReadBuffer
  JC Error_ReadSound
  LOOP Skiphd2
  CALL ReadBuffer
  JC Error_ReadSound
  CMP Buffer,1
  JNE Error_NotSupported
  MOV CX,17
Skiphd3:
  CALL ReadBuffer
  JC Error_ReadSound
  LOOP Skiphd3
  CALL ReadBuffer
  JC Error_ReadSound
  MOV AL,Buffer
  CALL ReadBuffer
  JC Error_ReadSound
  MOV AH,Buffer
  CALL ReadBuffer
  JC Error_ReadSound
  MOV DL,Buffer
  CALL ReadBuffer
  JC Error_ReadSound
  MOV DH,Buffer
  CMP DX,0
  JNE Error_SoundTooBig
  MOV SI,[BP+12]
  DEC SI
  SHL SI,1
  DEC AX
  MOV SoundLen[SI],AX
  SHL SI,1
  ADD SI,SoundPage
  MOV DX,SI
  PUSH AX
  PUSH BX
  CALL MapEMS
  POP BX
  MOV AX,EMSseg
  MOV DS,AX
  XOR SI,SI
  MOV EAX,080808080h
  MOV CX,16384
ClearSound:
  MOV [SI],EAX
  ADD SI,4
  LOOP ClearSound
  POP CX
  XOR DX,DX
  MOV AH,03Fh
  INT 21h
  JC Error_ReadSound
  MOV AX,03E00h
  INT 21h
  XOR AX,AX
  JMP EndLoadSound
Error_OpenSound:
  MOV AX,1          ; Error 1: Unable to open specified sound file
  JMP EndLoadSound
Error_ReadSound:
  MOV AX,03E00h
  INT 21h
  MOV AX,2          ; Error 2: General sound file read error
  JMP EndLoadSound
Error_BadSound:
  MOV AX,03E00h
  INT 21h
  MOV AX,3          ; Error 3: Bad sound file format
  JMP EndLoadSound
Error_NotSupported:
  MOV AX,03E00h
  INT 21h
  MOV AX,4          ; Error 4: Sound format not yet supported
  JMP EndLoadSound
Error_SoundTooBig:
  MOV AX,03E00h
  INT 21h
  MOV AX,5          ; Error 5: Sound file too big
  JMP EndLoadSound
Error_BadSlot:
  MOV AX,03E00h
  INT 21h
  MOV AX,6          ; Error 6: Sound slot does not exist
EndLoadSound:
  CALL UnhookInt24
  POP BP
  POP DS
  RET 6
xDQBloadSound ENDP

; ---------------------------------------------------------------------------
; DQBloadRawSound FUNCTION
; purpose:
;   Loads a sound sample into a specified sound slot. Sounds must be 8 bit
;   mono and their sampling rate must not be greater than 22000 Hz, although
;   this function does not check for the format to be supported.
;   DQBloadRawSound also requires the offset where the sound data begins into
;   the specified file, plus the sound length in bytes; in this way you can
;   store several sounds into the same file, and load them all using this
;   function.
; declaration:
;   DECLARE FUNCTION xDQBloadRawSound(BYVAL Slot,BYVAL FileSeg,BYVAL FileOff,
;                                     BYVAL HiOff,BYVAL LoOff,BYVAL Length)
;   DECLARE FUNCTION DQBloadRawSound(Slot AS INTEGER,FileName AS STRING,
;                                     Offset AS LONG,Length AS LONG)
; ---------------------------------------------------------------------------
PUBLIC xDQBloadRawSound
xDQBloadRawSound PROC
  ; Stack layout
  ;
  ; 18  Slot
  ; 16  FileSeg
  ; 14  FileOff
  ; 12  HiOff
  ; 10  LoOff
  ; 08  Length
  ; 06  Basic return segment
  ; 04  Basic return offset
  ; 02  DS
  ; 00  BP
  PUSH DS
  PUSH BP
  MOV BP,SP
  CALL HookInt24
  MOV AX,[BP+18]
  XOR BH,BH
  MOV BL,NumSounds
  CMP AX,BX
  JG Error_BadSlot1
  MOV AX,[BP+16]
  MOV DS,AX
  MOV DX,[BP+14]
  MOV AX,03D00h
  INT 21h
  JC Error_OpenRaw
  MOV BX,AX
  MOV AX,04200h
  MOV CX,[BP+12]
  MOV DX,[BP+10]
  INT 21h
  JC Error_ReadRaw
  MOV AX,@DATA
  MOV DS,AX
  MOV SI,[BP+18]
  DEC SI
  SHL SI,1
  MOV AX,[BP+8]
  MOV SoundLen[SI],AX
  SHL SI,1
  ADD SI,SoundPage
  MOV DX,SI
  PUSH BX
  CALL MapEMS
  POP BX
  MOV AX,EMSseg
  MOV ES,AX
  MOV DS,AX
  XOR DI,DI
  MOV EAX,080808080h
  MOV CX,16384
  REP STOSD
  MOV CX,[BP+8]
  MOV AH,03Fh
  XOR DX,DX
  INT 21h
  JC Error_ReadRaw
  CMP AX,0
  JE Error_EOF
  MOV AX,03E00h
  INT 21h
  XOR AX,AX
  JMP EndLoadRaw
Error_OpenRaw:
  MOV AX,1          ; Error 1: Unable to open specified sound file
  JMP EndLoadRaw
Error_ReadRaw:
  MOV AX,03E00h
  INT 21h
  MOV AX,2          ; Error 2: General sound file read error
  JMP EndLoadRaw
Error_EOF:
  MOV AX,03E00h
  INT 21h
  MOV AX,3          ; Error 3: Cannot read past the end of file
  JMP EndLoadRaw
Error_BadSlot1:
  MOV AX,4          ; Error 4: Specified sound slot does not exist
EndLoadRaw:
  CALL UnhookInt24
  POP BP
  POP DS
  RET 12
xDQBloadRawSound ENDP

; ---------------------------------------------------------------------------
; DQBplaySound SUB
; purpose:
;   Plays a sound previously loaded in memory
; declaration:
;   DECLARE SUB DQBplaySound(BYVAL SoundNum,BYVAL Voice,BYVAL LoopFlag)
; ---------------------------------------------------------------------------
PUBLIC DQBplaySound
DQBplaySound PROC
  ; Stack layout
  ;
  ; 10  SoundNum
  ; 08  Voice
  ; 06  LoopFlag
  ; 04  Basic return segment
  ; 02  Basic return offset
  ; 00  BP
  PUSH BP
  MOV BP,SP
  CMP SoundActive,0
  JE EndPlaySound
  MOV AX,[BP+10]
  XOR AH,AH
  CMP AL,NumSounds
  JG EndPlaySound
  MOV SI,[BP+8]
  DEC SI
  SHL SI,1
  CMP VoiceSound[SI],0
  JNE StillPlaying
  INC Playing
StillPlaying:
  MOV VoiceSound[SI],AX
  MOV AX,[BP+6]
  MOV VoiceLoop[SI],AX
  MOV VoicePos[SI],0
EndPlaySound:
  POP BP
  RET 6
DQBplaySound ENDP

; ---------------------------------------------------------------------------
; DQBpauseSound SUB
; purpose:
;   Pauses the samples sound output
; declaration:
;   DECLARE SUB DQBpauseSound()
; ---------------------------------------------------------------------------
PUBLIC DQBpauseSound
DQBpauseSound PROC
  CMP PauseSound,1
  JE EndPauseSound
  MOV DX,BaseAddr
  ADD DX,0Ch
  WAITWRITEDSP
  MOV AL,0D0h
  OUT DX,AL
  MOV PauseSound,1
EndPauseSound:
  RET
DQBpauseSound ENDP

; ---------------------------------------------------------------------------
; DQBresumeSound SUB
; purpose:
;   Resumes the samples sound output
; declaration:
;   DECLARE SUB DQBresumeSound()
; ---------------------------------------------------------------------------
PUBLIC DQBresumeSound
DQBresumeSound PROC
  CMP PauseSound,0
  JE EndResumeSound
  MOV DX,BaseAddr
  ADD DX,0Ch
  WAITWRITEDSP
  MOV AL,0D4h
  OUT DX,AL
  MOV PauseSound,0
EndResumeSound:
  RET
DQBresumeSound ENDP

; ---------------------------------------------------------------------------
; DQBstopVoice SUB
; purpose:
;   Stops sound playing on specified voice
; declaration:
;   DECLARE SUB DQBstopVoice(BYVAL Voice)
; ---------------------------------------------------------------------------
PUBLIC DQBstopVoice
DQBstopVoice PROC
  PUSH BP
  MOV BP,SP
  MOV SI,[BP+6]
  DEC SI
  SHL SI,1
  CMP VoiceSound[SI],0
  JE EndStopVoice
  DEC Playing
  MOV VoiceSound[SI],0
EndStopVoice:
  POP BP
  RET 2
DQBstopVoice ENDP

; ---------------------------------------------------------------------------
; DQBremoveSB SUB
; purpose:
;   Turns off SB output
; declaration:
;   DECLARE SUB DQBremoveSB()
; ---------------------------------------------------------------------------
PUBLIC DQBremoveSB
DQBremoveSB PROC
  CMP SoundActive,0
  JE EndRemoveSB
  MOV DX,BaseAddr
  ADD DX,06h
  MOV AL,1
  OUT DX,AL
  IN AL,DX
  IN AL,DX
  IN AL,DX
  IN AL,DX
  XOR AL,AL
  OUT DX,AL
  ADD DX,08h
  WAITREADDSP
  SUB DX,4
  IN AL,DX          ; SB reset
  IN AL,021h
  OR AL,PicORmask
  OUT 021h,AL
  XOR AX,AX
  MOV ES,AX
  MOV AL,IRQ
  ADD AL,8
  MOV SI,AX
  SHL SI,2
  CLI
  MOV AX,OldIRQoff
  MOV ES:[SI],AX
  MOV AX,OldIRQseg
  MOV ES:[SI+2],AX  ; Old IRQ restored
  STI
  MOV SoundActive,0
EndRemoveSB:
  RET
DQBremoveSB ENDP

; ---------------------------------------------------------------------------
; DQBsetVolume SUB
; purpose:
;   Sets the master volume for sound output, in the range 0-15
; declaration:
;   DECLARE SUB DQBsetVolume(BYVAL Volume)
; ---------------------------------------------------------------------------
PUBLIC DQBsetVolume
DQBsetVolume PROC
  PUSH BP
  MOV BP,SP
  MOV DX,BaseAddr
  ADD DX,4
  MOV AL,022h
  OUT DX,AL
  INC DX
  MOV AX,[BP+6]
  AND AL,0Fh
  MOV BL,AL
  SHL BL,4
  OR AL,BL
  OUT DX,AL
  POP BP
  RET 2
DQBsetVolume ENDP


END
